/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *   gfx.cpp - contains all graphic operations that could be used, namely  *
 *             single pixel operations, color multiplications, flood fill  *
 *             and even sphere pre-filler to speed the things up.          *
 *                                                                         *
 ***************************************************************************/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "MyGlobal.h"
#include "MyTypes.h"
#include "antialias.h"
#include "gfx.h"
#include "vectormath.h"

// this is 1/3.5
#define CRITSTR_UPDATE 0.2857
#define QUEUESIZE 1200
#define M_SIDE_CONSTANT 6.6666

const double nocalcthresh = 10;
const double CircConst    = 0.0000390625;

double lastgX=-9999999.0, lastgY=-9999999.0, lastr;

// default resolution:
int Xres=640, Yres=480;
int Xres2, Yres2;
extern int sse_enabled;
extern int camera_moved;
extern int RowMin[], RowMax[];


// PreSC[i][j][0] = sin(j/i*2*pi)
// PreSC[i][j][1] = cos(j/i*2*pi)
double PreSC[MAX_SPHERE_SIDES+1][MAX_SPHERE_SIDES+1][2];

#ifdef ACTUALLYDISPLAY
void Slock(SDL_Surface *screen)
{
  if ( SDL_MUSTLOCK(screen) )
  {
    if ( SDL_LockSurface(screen) < 0 )
    {
      return;
    }
  }
}

void Sulock(SDL_Surface *screen)
{
  if ( SDL_MUSTLOCK(screen) )
  {
    SDL_UnlockSurface(screen);
  }
}
#endif

int xres(void){
	return Xres;
}

int yres(void){
	return Yres;
}

void SetNewMode(int x, int y)
{
 if (x%RESOLUTION_X_ALIGN) {
 	printf("I can't set the desired resolution (%dx%d)\n - setting X to the previous multiple-of-%d value\n",
	x, y, RESOLUTION_X_ALIGN);
	x&=~(RESOLUTION_X_ALIGN-1);
 }
 y = x/4*3;
 Xres = x;
 Yres = y;
 Xres2 = xsize_render(x);
 Yres2 = ysize_render(y);
 if (Xres2 > RES_MAXX || Yres2 > RES_MAXY) {
 	set_fsaa_mode(FSAA_MODE_NONE);
 	Xres2 = xsize_render(x);
 	Yres2 = ysize_render(y);
 	}
}

void gfx_update_2nds(void)
{
 Xres2 = xsize_render(Xres);
 Yres2 = ysize_render(Yres);
}

void DrawPixel(Uint32 *p, int x, int y, Uint8 r, Uint8 g, Uint8 b)
{
 Uint32 *bfp;
 bfp = p + y*Xres + x;
 *bfp = (int) b + (((int)g)<<8) + (((int)r)<<16);
}

void DrawPixeld(Uint32 *p, int x, int y, Uint32 v)
{
 Uint32 *bfp;
 bfp = p + y*Xres + x;
 *bfp = v;
}

Uint32 GetPixeld640(Uint32 *p, int x, int y)
{
 Uint32 *bfp;
 bfp = p + y*640 + x;
 return ((*bfp));
}

void DrawPixel640(Uint32 *p, int x, int y, Uint8 r, Uint8 g, Uint8 b)
{
 Uint32 *bfp;
 bfp = p + y*640 + x;
 *bfp = (int) b + (((int)g)<<8) + (((int)r)<<16);
}

void DrawPixeldsafe(Uint32 *p, int x, int y, Uint32 v)
{
 Uint32 *bfp;
 if ((unsigned) x >= (unsigned)Xres | (unsigned) y >= (unsigned)Yres)
 	return;
 bfp = p + y*Xres + x;
 *bfp = v;
}

void DrawPixeldxsafe(Uint32 *p, int x, int y, Uint32 v)
{
 Uint32 *bfp;
 if ((unsigned) x >= (unsigned)Xres2 | (unsigned) y >= (unsigned)Yres2)
 	return;
 bfp = p + y*Xres2 + x;
 *bfp = ((*bfp)<<16) + v;
}


Uint32 GetPixeld(Uint32 *p, int x, int y)
{
 Uint32 *bfp;
 bfp = p + y*Xres + x;
 return (*bfp);
}

void DrawPixeldx(Uint32 *p, int x, int y, Uint32 v)
{
 Uint32 *bfp;
 bfp = p + y*Xres2 + x;
 *bfp = ((*bfp)<<16) + v;
}


Uint32 GetPixeldx(Uint32 *p, int x, int y)
{
 Uint32 *bfp;
 bfp = p + y*Xres2 + x;
 return ((*bfp)&0xffff);
}


void LineFast( Uint32 *p, int x0, int y0, int x1, int y1, Uint32 color)
{int x, y, i, steps, xi, yi, x_int, y_int;

 x = abs(x1-x0); y = abs(y1-y0);
 if (x>y) steps = x; else steps = y;
 if (!steps) steps =1 ;

 xi = (int) (65536.0 * ((double)(x1-x0)/(double)steps));
 yi = (int) (65536.0 * ((double)(y1-y0)/(double)steps));
 x = x0 << 16;
 y = y0 << 16;
 for (i=1;i<=steps;i++) {
 	x_int = x>>16;
	y_int = y>>16;
	if ((unsigned) y_int < (unsigned) Yres2) {
	//if (y_int >= 0 && y_int < Yres2) {
 		DrawPixeldxsafe(p, x_int, y_int, color);
		if (x_int > RowMax[y_int])
			RowMax[y_int] = x_int;
		if (x_int < RowMin[y_int])
			RowMin[y_int] = x_int;
	}
 	x += xi;
	y += yi;
 }
}




// draws transparent pixel at X, Y
// with color col and transparency factor alpha
void DrawTransparent(Uint32 *p, int x, int y, Uint32 col, float alpha)
{Uint32 z;
 if ((unsigned)x>=(unsigned)Xres||(unsigned)y>=(unsigned)Yres) return;
 z = GetPixeld(p, x, y);
 DrawPixel(p, x, y, (int) (alpha * ((col>>16)& 0xff) + (1-alpha) * ((z >> 16)& 0xff)),
                    (int) (alpha * ((col>>8)& 0xff) + (1-alpha) * ((z >> 8)& 0xff)),
		    (int) (alpha * (float) (col&0xff) + (1-alpha) * (float) (z& 0xff)));
}

void wLine(Uint32 *p, int x1, int y1, int x2, int y2, Uint32 col)
{int i, steps, v1, v2;
 float x, xi=0, y, yi=0;
 v1 = x1-x2;
 v2 = y1-y2;
 if (v1<0) v1 = -v1;
 if (v2<0) v2 = -v2;
 steps = (v1>v2?v1:v2);
 if (steps) {
    xi = (x2-x1)/(float)steps;
    yi = (y2-y1)/(float)steps;
    }
 if (v1<v2) {
    for (i=0,x=x1,y=y1;i<=steps;i++,x+=xi,y+=yi) {
	DrawTransparent(p, (int) floor(x), (int) y, col, ceil(x)-x);
	DrawTransparent(p, (int) ceil(x), (int) y, col, x-floor(x));
	}
    } else
    {
    for (i=0,x=x1,y=y1;i<=steps;i++,y+=yi,x+=xi) {
	DrawTransparent(p, (int) x, (int) floor(y), col, ceil(y)-y);
	DrawTransparent(p, (int) x, (int) ceil (y), col,y-floor(y));
	}
    }
}

#ifdef ACTUALLYDISPLAY
/* PasteRaw -- puts a RawImg on a surface at coords x, y */
/* comments: really FAST */
void PasteRaw(SDL_Surface *p, RawImg a, int x, int y)
{int i, bytx=a.x*4, rowy=a.y;
 Uint32 *xx ;
 if (y+a.y>p->h)
 	rowy = p->h - y;
 if (x+a.x>p->w)
 	bytx = (p->w - x)*4;
 Slock(p);
 for (i=0;i<rowy;i++)
 	{
	xx = (Uint32 *) p->pixels + y*p->pitch/4 + i*p->pitch/4 + x;
 	memcpy(xx, &a.data[i*a.x], bytx);
	}
 Sulock(p);
 SDL_Flip(p);
}

/* Shows a progress bar (for the intro purposes), with a progress prog (should be within 0.0 and 1.0)
   flips every time ;( try to use as rarely as possible ) */
void intro_progress(SDL_Surface *p, double prog)
{
 int i, j, endx, sx=31;
 int rg = 81, b = 250, rgi=2, bi=4;
 if (prog>1) prog = 1.0;
 Slock(p);
 endx = (int) (31 + 592.0 * prog);
 for (j=233;j<=247;j++, sx--, endx--, rg-=rgi, b-=bi) {
	for (i=sx;i<=endx;i++)
		if (GetPixeld640((Uint32*)p->pixels, i, j) < 0x300000) {
			DrawPixel640((Uint32*)p->pixels, i, j, rg + rand()%5-2, rg + rand()%5-2, b + rand()%10-5);
			}
 	}
 Sulock(p);
 SDL_Flip(p);
}
#endif

#ifdef USE_ASSEMBLY
SSE_ALIGN(static float w2_flones[4]) = {1.0, 1.0, 1.0, 1.0};
#endif

#define gfx_asm
#include "x86_asm.h"
#undef gfx_asm

Uint32 blend_p5(Uint32 foreground, Uint32 background, float opacity)
{
	return(multiplycolorf(foreground, opacity) + multiplycolorf(background, (1-opacity)));
}

// blends two colors : more opacity => result is closer to foreground.
// uses assembly SSE version if it is supported
Uint32 blend(Uint32 foreground, Uint32 background, float opacity)
{
 if ( sse_enabled )	return(blend_sse(foreground, background, opacity));
 		else	return(multiplycolorf(foreground, opacity) + multiplycolorf(background, (1-opacity)));
}

// amplifies color by amp (amp=65536 (=1.0 in float) -> no change)
Uint32 multiplycolor(Uint32 color, int amp)
{int r, g, b;
 b = (amp * (color & 0xff))>>16; color>>=8;
 g = (amp * (color & 0xff))>>16; color>>=8;
 r = (amp * (color & 0xff))>>16;
 b = (b>255?255:b); g = (g>255?255:g); r = (r>255?255:r);
 return (b + (g << 8) + (r << 16));
}

// amplifies color by amp (amp=1.0 -> no change)
Uint32 multiplycolorf(Uint32 color, float amp)
{int r, g, b;
 b = (int) ((amp * (double)(color & 0xff))); color>>=8;
 g = (int) ((amp * (double)(color & 0xff))); color>>=8;
 r = (int) ((amp * (double)(color & 0xff)));
 r = (r>255?255:r); g = (g>255?255:g); b = (b>255?255:b);
 return (b + (g << 8) + (r << 16));
}

// returns the given color, desaturated to factor (0 - no desaturation. 1 - all color info is lost)
Uint32 desaturate(Uint32 color, double factor)
{int r, g, b,m;
 b = color & 0xff; color>>=8;
 g = color & 0xff; color>>=8;
 r = color & 0xff;
 m = (r+g+b)/3;
 b = b + (int)(factor * ((double)(m-b)));
 g = g + (int)(factor * ((double)(m-g)));
 r = r + (int)(factor * ((double)(m-r)));
 return (b | (g << 8) |(r<<16));
}

/*
	The following idea of a fast reciprocal of a 4th root approximation is influenced directly from
	http://playstation2-linux.com/files/p2lsd/fastrsqrt.pdf

	In fact, the algorithm suggested ther worked fine for 4th root (instead of square root) with a simple
	constants substitution ...

	10x to Matthew Jones for this amazing piece of work.

	(	the function approximates 1/sqrt(sqrt(x)).	The error is within +/-3 % 	)

	I have *NO* idea how this thing works
*/
#define uint unsigned int
static inline float FastPower_minus_quarter(float val)
{
    const float magicValue = 1331185664.0;
    float tmp = (float) *((uint*)&val);
    tmp = (tmp * -0.25) + magicValue;
    uint tmp2 = (uint) tmp;
    return *(float*)&tmp2;
}

// calculates the formula darkening which should be due to the distance from the light source
float lightformulae_tiny(float x)
{
	return 25.0 * FastPower_minus_quarter(x);
}

static inline Uint32 addrgb(Uint32 A, Uint32 B)
{int r,g,b;

 b = (A&0xff) + (B&0xff);// b = (b>255?255:b);
 g = ((A>>8)&0xff) + ((B>>8)&0xff); //g = (g>255?255:g);
 r = ((A>>16)&0xff) + ((B>>16)&0xff); //r = (r>255?255:r);

 return ((r<<16)+(g<<8) + b);

}

#ifdef ACTUALLYDISPLAY
// a printf-like procedure... prints a message at x,y with a color col and the given opacity
void printxy(SDL_Surface *p, Uint32 *a, RawImg font, int x, int y, Uint32 col, float opacity, const char *buf, ...)
{
 va_list arg;
 static char s[1024];
 int i, fx, x0, y0;
 int ampi, indeh;
 double amp;

 va_start(arg, buf);
#ifdef _MSC_VER
 _vsnprintf(s, 1024, buf, arg); // M$ sucks with those _names :)
#else
 vsnprintf(s, 1024, buf, arg);
#endif
 va_end(arg);

 i = 0;
 while (s[i] && i < 1024) {
	if (s[i]>32 && s[i]<126) {
		fx = (int) ((s[i]-33)*FONT_XSIZE);
		for (x0=0;x0<FONT_XSIZE_FLOOR; x0++)
			for (y0=0;y0<FONT_YSIZE;y0++) {
				ampi= font.data[fx+x0 + y0*font.x] & 0xff;
				amp = ampi / 255.0;
				ampi<<=8;
				if (amp>0.0005) {
					amp *= opacity;
					indeh = x + x0 + (y + y0)*Xres;
					a[ indeh ] = addrgb	(
									multiplycolor(a[ indeh ], 65535 - ampi),
									multiplycolor(col, ampi)
								);
					}
				}
		}
 	i++; x+=FONT_XSIZE_CEIL;
	if (x>=Xres-1- FONT_XSIZE_CEIL) break;
 	}
}
#endif

static inline int inrange(int x, int y)
{
 return ((unsigned) x< (unsigned)Xres && (unsigned) y<(unsigned)Yres);
}

static inline int inrange2(int x, int y)
{
 return ((unsigned) x<(unsigned)Xres2 && (unsigned) y<(unsigned)Yres2);
}


void FloodFill(Uint32 *p, int x, int y, Uint32 c)
{int a[QUEUESIZE][2], b[QUEUESIZE][2];
 int ac, bc, i, nx, ny;
 if (!inrange2 (x, y) || GetPixeldx(p, x, y)==c) return;
 ac=1; a[0][0]=x; a[0][1]=y; DrawPixeldx(p, x, y, c);

 while (ac) {
 	bc = 0;
	for (i=0;i<ac;i++) {
	nx = a[i][0]+1; ny=a[i][1];
	if (inrange2(nx, ny) && GetPixeldx(p, nx, ny)!=c) {
		DrawPixeldx(p, nx, ny, c);
		b[bc][0]=nx;
		b[bc++][1]=ny;
		}
	nx = a[i][0]-1; ny=a[i][1];
	if (inrange2(nx, ny) && GetPixeldx(p, nx, ny)!=c) {
		DrawPixeldx(p, nx, ny, c);
		b[bc][0]=nx;
		b[bc++][1]=ny;
		}
	nx = a[i][0]; ny=a[i][1]+1;
	if (inrange2(nx, ny) && GetPixeldx(p, nx, ny)!=c) {
		DrawPixeldx(p, nx, ny, c);
		b[bc][0]=nx;
		b[bc++][1]=ny;
		}
	nx = a[i][0]; ny=a[i][1]-1;
	if (inrange2(nx, ny) && GetPixeldx(p, nx, ny)!=c) {
		DrawPixeldx(p, nx, ny, c);
		b[bc][0]=nx;
		b[bc++][1]=ny;
		}
	if (bc>QUEUESIZE-5) return ;
		}
	ac=bc;
	for (i=0;i<ac;i++) { a[i][0]=b[i][0]; a[i][1]=b[i][1]; }
 	}

}


/* calculates the needed number of "shell" 3D points needed for the pre-filler

	In fact, it generates points of the circle, which is the cross bween the given sphere and the plane A' (perpendicular to
	the ray CA.) It determines how many points to generate so that the resulting polygon would be smooth enough.

*/
int calculate_convex(sphere *a, double pt[][3], double *c)
{double alpha, beta;
 double sina, sinb, cosa, cosb;
 int steps,i;
 double *f;
 //printf("calculating convex: ");
 //steps = (int) (a->d * Xres / a->dist * M_SIDE_CONSTANT);
 steps = 4;
 while (steps < MAX_SPHERE_SIDES &&
 	a->d*sqrt(2.0f-2.0f*cos(2.0f*M_PI/steps)) / a->dist * Xres > M_SIDE_CONSTANT)
		steps+=4;
 if (steps == MAX_SPHERE_SIDES) steps-=4;
 if (steps>=MAX_SPHERE_SIDES) steps = MAX_SPHERE_SIDES-1;
 if (steps<MIN_SPHERE_SIDES) steps = MIN_SPHERE_SIDES;
 //printf("Steps=%d ", steps);

 f = &PreSC[steps][0][0];

 alpha = atan2(a->x-c[0], a->z-c[2])+M_PI/2;
 beta = atan2(a->y - c[1], sqrt(sqr(a->z-c[2]) + sqr(a->x-c[0])));
 //printf("alpha=%.2lf, beta=%.2lf ", alpha, beta);
 sina = sin(alpha);
 sinb = sin(beta);
 cosa = cos(alpha);
 cosb = cos(beta);

 for (i=0;i<=steps;i++) {
 	pt[i][0] = a->x + a->d * (sina * f[1] + cosa * f[0] * sinb);
	pt[i][1] = a->y + a->d *  f[0] * cosb;
	pt[i][2] = a->z + a->d * (cosa * f[1] - sina * f[0] * sinb);
 	f+=2;
 	}
 return steps;
}

static int xdev[4] = { 0, +1,  0, -1};
static int ydev[4] = {+1,  0, -1,  0};

/*
	Maps a sphere on the framebuffer. Pixels, coloured with its number are "expected" to be rendered there
	The Mapper's error is within 0.4% in most cases.
*/
void mapsphere(Uint32 *fb, int Ox, int Oy, int color, int sides, double pt[][3], double *c, double w[3][3], sphere *A)
{int i, x0, y0, x1, y1;
 int xs[4], ys[4], sc=0, waited;

 waited = sides / 4;

 ProjectPoint(&x0, &y0, pt[0], c, w, Xres2, Yres2);
 xs[0] = x0; ys[0] = y0;
 for (i=1;i<=sides; i++) {
 	ProjectPoint(&x1, &y1, pt[i], c, w, Xres2, Yres2);
	//DrawPixeldsafe(fb, x, y, color);
	LineFast(fb, x0, y0, x1, y1, color);
	x0 = x1; y0 = y1;
	if ( i == waited ) {
		waited += sides / 4;
		xs[sc] = x1; ys[sc++] = y1;
		if (sc == 4) waited = -1;
		}
 	}
 if (A->d/A->dist*Xres>2)
 	if (inrange2(Ox, Oy))
 		FloodFill(fb, Ox, Oy, color);
		else{
			for (i=0;i<4;i++) {
				x0 = xs[i] + xdev[i]; y0 = ys[i] + ydev[i];
				if (inrange2(x0, y0)) {
					FloodFill(fb, x0, y0, color);
					break;
				}
			}
		}
}

// the pre-filler part: this fills the circular area of the screen where the selected sphere will map with the given color
// returns 0 if the sphere does not visualise at all
int projectIt(sphere *a, double pt[][3], int *ns, Uint32 *fb, double *cur, double w[3][3], int xres, int yres, int color)
{int x, y;
 if (!ProjectSphere(a, cur, w, &x, &y, xres, yres)) return 0;
 //mapsphere(fb, x, y, a.d/a.dist*(xres/2), atan2(x-(xres/2), y-(yres/2)), sqrt(sqr(x/(double) xres - 0.5) + sqr(y / (double)yres - 0.5)), color,1);
 //if (!inrange(x, y)) return 0;
 if (camera_moved)
 	*ns = calculate_convex(a, pt, cur);
 mapsphere(fb, x, y, color, *ns, pt, cur, w, a);
 return 1;
}

void gfx_init(void)
{
 int i,j;

 for (j=1;j<=MAX_SPHERE_SIDES;j++)
 	for (i=0;i<=j;i++) {
		PreSC[j][i][0] = sin(((double)i/(double)j)*2*M_PI);
		PreSC[j][i][1] = cos(((double)i/(double)j)*2*M_PI);
		}
 gfx_update_2nds();
}

void gfx_close(void)
{
}
