/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *  render.cpp - all rendering routines, seperated for non-SSE and SSE-    *
 *               enabled CPU's                                             *
 *                                                                         *
 ***************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "MyGlobal.h"
#include "MyTypes.h"
#include "antialias.h"
#include "bitmap.h"
#include "cmdline.h"
#include "fract.h"
#include "gfx.h"
#include "profile.h"
#include "render.h"
#include "rgb2yuv.h"
#include "shaders.h"
#include "sphere.h"
#include "vectormath.h"

#include "cross_vars.h"

#ifdef ACTUALLYDISPLAY
#include <SDL/SDL_thread.h>
#else
#include <pthread.h>
#include <semaphore.h>
#endif

#define fract_asm
#include "x86_asm.h"
#undef fract_asm

//Uint32 savebuffer[20*30][400*300]; // 20 s at 30 fps @ 400x300

SSE_ALIGN(Uint32 floorbuffer[RES_MAXX*RES_MAXY]);
SSE_ALIGN(unsigned short fbuffer[RES_MAXX*RES_MAXY]);
SSE_ALIGN(unsigned short sbuffer[RES_MAXX*RES_MAXY]);
int RowInfo[RES_MAXY];
int RowMax[RES_MAXY], RowMin[RES_MAXY];
double Area_const = 1.95;
int UseThreads = -1;
int do_mipmap  = 1;
int r_shadows  = 0;
int cpu_count = 1;

#ifdef FSTAT
static int good, worse, realbad;
#endif


/* integrated lighting and Multiplication routine.
 * this routine replaces the following code:
 *
 *	screencolor = multiplycolor(texturecolor, lform(Lx, Ly, lx, ly))
 * being used somewhere in DrawIt.
 *
 * uses an integrated assembly code in x86_asm.h (using SSE) if the required instruction set is available
 */
Uint32 integrated( Uint32 texturecolor, int x, int y, int lx, int ly)
{
 if ( sse_enabled )  return igrated(texturecolor, x, y, lx, ly, ysqrd_raytrace);
 		else return multiplycolor(texturecolor, lform(x, y, lx, ly, ysqrd_raytrace));
}

/* This is called by the raytracer - gets a texture color at coords (x, 0, y) */

Uint32 texture_handle_nearest(int x, int y)
{
 return integrated(T[TEX_S].data[((x & 255)>>TEX_S) +
 /*y*/(((y & 255)>>TEX_S)<<(LOG2_MAX_TEXTURE_SIZE-TEX_S))],x, y, lx, lz);
}

Uint32 texture_handle_bilinear(double x, double y)
{
 Uint32 x1y1, x2y1, x1y2, x2y2;
 int off, mask;
 int xi = (int)x;
 int yi = (int)y;
        /*x*/               /*y*/
 off = ((xi & 255)>>TEX_S) + (((yi & 255)>>TEX_S)<<(LOG2_MAX_TEXTURE_SIZE-TEX_S));
 mask = (1 << (2*(LOG2_MAX_TEXTURE_SIZE-TEX_S))) - 1;
 //mask = 1023;
 x1y1 = T[TEX_S].data[off];
 x2y1 = T[TEX_S].data[(off+1)&mask];
 off += (1 << (LOG2_MAX_TEXTURE_SIZE-TEX_S));
 x1y2 = T[TEX_S].data[off & mask];
 x2y2 = T[TEX_S].data[(off+1)&mask];
 x/=(1<<TEX_S); y/=(1<<TEX_S);
 x-=floor(x); y-=floor(y);
 if (sse_enabled) {
 	return igrated(bilinea4(x1y1, x2y1, x1y2, x2y2, (int) (65536.0*x), (int) (65536.0*y)), xi, yi, lx, lz, ysqrd_raytrace);
	} else {
	return  multiplycolor(
		multiplycolorf(	x1y1, (1-x)*(1-y))+
		multiplycolorf(	x2y1,    x *(1-y))+
		multiplycolorf(	x1y2, (1-x)*   y )+
		multiplycolorf(	x2y2,    x *   y )
			, lform(xi, yi, lx, lz, ysqrd_raytrace));
	}
}

/*
function bilinea_p5:

does bilinear filtering on the four points (x0y0, x1y0, x0y1 and x1y1) using Fx and Fy as .16 fixedpoint represenation
of the coordinates. Basically, the routine does what the equivallent code from render.cpp did in versions prior to 1.03c:

		Tc = multiplycolor(T[xtex].data[boffset				], i2im(65536-Fx,65536-Fy))+
		multiplycolor(T[xtex].data[(boffset+1) &texandmask		], i2im(Fx  ,65536-Fy))+
		multiplycolor(T[xtex].data[(boffset+xandmask+1)&texandmask	], i2im(65536-Fx,  Fy))+
		multiplycolor(T[xtex].data[(boffset+xandmask+2)&texandmask	], i2im(Fx  ,  Fy));

	the speedup comes from the fact, that instead of 4 unpacks and 4 repacks(as in the above version),
	only one repack is done. Also, it attempts to exploit parallelism on the CPU's that have deep pipelines,
	but the only pre-SSE CPUs which could have pipelines that deep are maybe just the AMD K7s (T-Birds and Spitfires)
*/
static inline Uint32 bilinea_p5(Uint32 x0y0, Uint32 x1y0, Uint32 x0y1, Uint32 x1y1, unsigned Fx, unsigned Fy)
{
	unsigned m0, m1, m2, m3;
	int r0, g0, b0, r1, g1, b1;
	Uint32 rez=0;

	if (mmx_enabled) {
	/* do we have MMX? If we do, we may use the crappy MMX bilinear filter, which, due to lack of
	   the pshufw instruction is kinda slow...
	*/
		Fx>>=8; Fy>>=8;
//#define fract_asm
#define bilinea_p5_asm
#include "x86_asm.h"
#undef bilinea_p5_asm
//#undef fract_asm
		return rez;
	}
// non-mmx version
	m1 = m0 = 65536 - Fy;
	m2 = 65536 - (m3=Fx);
	m0 *= m2;
	m1 *= Fx;
	m2 *= Fy;
	m3 *= Fy;
	m0 >>= 16; m1>>=16; m2 >>= 16; m3>>=16;

	b0  = m0 * (x0y0&0xff); x0y0 >>= 8;
	g0  = m0 * (x0y0&0xff); x0y0 >>= 8;
	r0  = m0 * (x0y0&0xff);
	b1  = m1 * (x1y0&0xff); x1y0 >>= 8;
	g1  = m1 * (x1y0&0xff); x1y0 >>= 8;
	r1  = m1 * (x1y0&0xff);

	b0 += m2 * (x0y1&0xff); x0y1 >>= 8;
	g0 += m2 * (x0y1&0xff); x0y1 >>= 8;
	r0 += m2 * (x0y1&0xff);
	b1 += m3 * (x1y1&0xff); x1y1 >>= 8;
	g1 += m3 * (x1y1&0xff); x1y1 >>= 8;
	r1 += m3 * (x1y1&0xff);


	b0 += b1; g0 += g1; r0 += r1;
	b0 >>= 16; g0>>=16; r0>>=16;
	return (b0 + (g0 << 8) + (r0 << 16));
}



/*
 *	This handles camera movement using current direction vector of the camera and the supplied
 *	`angle' param (angle = 0 means "move straight ahead"; angle = -M_PI means 'move backwards' and so on.
 */
void move_camera(double angle)
{double a, b, x[3];
 int i;
 a = alpha + angle;
 b = beta;
 if (angle == M_PI/2 || angle == -M_PI/2) b = 0;
 if (angle == M_PI)	b = -b;
 x[0] = sin(a)*cos(b);
 x[1] = sin(b);
 x[2] = cos(a)*cos(b);
 for (i=0;i<3;i++)
 	cur[i]+=(x[i]*delta*speed);
 camera_moved = 1;
}

/*
 *	Rotates the camera vector, depending on last frame generation time and mouse sensitivity
 */
void rotate_camera(double alphai, double betai)
{
 alpha += (alphai*delta*anglespeed);
 beta +=  (betai *delta*anglespeed);
}

void check_params(void)
{
 if (cur[1]<daFloor+1) cur[1] = daFloor+1;
 if (cur[1]>daCeiling -1) cur[1] = daCeiling -1;
 if (beta<-M_PI/2) beta = -M_PI/2;
 if (beta> M_PI/2) beta =  M_PI/2;
}

double czero(double x)
{
 if (x<0.0001) x = 9999;
 return x;
}

Uint32 *get_image_buffer(void)
{
	if ( sse_enabled )
		return floorbuffer;
		else return framebuffer;
}

/*
 *	This is the function to cope with the blur effect (if it is enabled in MyGlobal.h) - the idea used here is
 *	to generate 8 consecutive frames and merge them in a single frame which is copied to the real framebuffer and
 *	displayed as usual. We are using src for temporary (hidden) framebuffer, `bb' to accumulate result from the
 *	eight frames. `fb' is the `real' framebuffer. Note that pixels are right shifted by 3 before added to the
 *	accumulation buffer (this does serious image degradation, since output image is 15 bit).
 *
 *	A MMX version of this routine exists in x86_asm.h
 */

void blur_do(Uint32 *bb, Uint32 *src, Uint32 *fb, int count, int shouldcopy)
{
	int i;
	if (mmx_enabled)
		blur_do_MMX(bb, src, fb, count, shouldcopy);
		else {
			for(i=0;i<xres()*yres();i++)
				bb[i]+= ((src[i] >> 3) & 0x1f1f1f1f);
			if (shouldcopy) {
				memcpy(fb, bb, count*4);
				for (i=0;i<count;i++)
					bb[i]=0;
 			}

		}
}
/*
 *	This code deals with post-frame generation stuff, which is common for the P5 and SSE version of DrawIt
 */
#ifdef ACTUALLYDISPLAY
void postframe_do(SDL_Surface *p, SDL_Overlay *ov, Uint32 *fb)
#else
void postframe_do(Uint32 *fb)
#endif
{
 static char buff[1024];
 Uint32 *sfb;
#ifdef ACTUALLYDISPLAY
 SDL_Rect ddestrect;

 ddestrect.x = 0;
 ddestrect.y = 0;
 ddestrect.w = xres();
 ddestrect.h = yres();
#endif
	prof_enter(PROF_ANTIALIAS);
	sfb = fb;
#ifdef BLUR
 antialias(blurbuffer);
#else
 antialias(fb);
#endif
	prof_leave(PROF_ANTIALIAS);
	fb = sfb;
	//memcpy(savebuffer[vframe+1], sfb, 400*300*4);
	prof_enter(PROF_SHADER);
#ifdef BLUR
    if (use_shader)
 	shader_edge_glow(blurbuffer, blurbuffer, xres(), yres());
#else
    if (use_shader)
	//shader_inversion(fb, fb, xres(), yres(), xres()/5);
	shader_edge_glow(fb, fb, xres(), yres());
	//shader_blur(fb, fb, xres(), yres(), 7);
	//shader_FFT_Filter(fb, fb, xres(), yres());
#endif
 	prof_leave(PROF_SHADER);
#ifdef MAKE_CHECKSUM
	prof_enter(PROF_CHECKSUM);
 for (i=0;i<xres()*yres();i++) {
 	cksum += (fb[i]&0xffff)^(fb[i]>>16);
	}
	prof_leave(PROF_CHECKSUM);
#endif
#ifdef BLUR
	prof_enter(PROF_BLUR_DO);
 blur_do(bb1, blurbuffer, fb, xres()*yres(), (vframe%8==0));
 	prof_leave(PROF_BLUR_DO);
#endif
#ifdef ACTUALLYDISPLAY
#ifdef SHOWFPS
	prof_enter(PROF_SHOWFPS);
 if (developer) {
 	fpsfrm++;
 	if (GetTicks()-fpsclk>FPS_UPDATE_INTERVAL || not_fps_written_yet) {
 		not_fps_written_yet = 0;
 		sprintf(buff, "fps: %.1lf", (double) fpsfrm / ((GetTicks()-fpsclk)/1000.0));
		printxy(p, fb, font0, 0, 0, FPS_COLOR, 0.8, buff);
 		fpsfrm = 0;
 		fpsclk = GetTicks();
 		} else printxy(p, fb, font0, 0, 0, FPS_COLOR, 0.8, buff);
 }
 	prof_leave(PROF_SHOWFPS);
#endif // SHOWFPS
#endif // ACTUALLYDISPLAY
#ifdef ACTUALLYDISPLAY
 if (ov==NULL) { // use the surface directly
 	Slock(p);
		prof_enter(PROF_MEMCPY);
 	memcpy(p->pixels, fb, p->h*p->w*4);
		prof_leave(PROF_MEMCPY);
 	Sulock(p);
		prof_enter(PROF_SDL_FLIP);
 	SDL_Flip(p);
		prof_leave(PROF_SDL_FLIP);
 	}
	else { // use the overlay
	SDL_LockYUVOverlay(ov);
		prof_enter(PROF_CONVERTRGB2YUV);
	ConvertRGB2YUV((Uint32 *) ov->pixels[0], fb, p->h*p->w);
		prof_leave(PROF_CONVERTRGB2YUV);
	SDL_UnlockYUVOverlay(ov);
		prof_enter(PROF_DISPLAYOVERLAY);
	SDL_DisplayYUVOverlay(ov, &ddestrect);
		prof_leave(PROF_DISPLAYOVERLAY);
	}
#endif // ACTUALLYDISPLAY
 vframe++;
#ifdef FSTAT
 printf("frame %6d: good/bad/realbad = %7.3lf%%/%7.3lf%%/%7.3lf%%\n", vframe,
 (double) 100.0*good    / ((double) xres()*yres()),
 (double) 100.0*worse   / ((double) xres()*yres()),
 (double) 100.0*realbad / ((double) xres()*yres())); fflush(stdout);
#endif
}

/*
 *	This routine does the common stuff, identical for both the P5 and SSE version of Drawit
 */

void preframe_do(Uint32 *ptr, double *lw)
{
 int i, j;
 double btm, ttm;
 double obex[3];

 ysqrd_ceil = (int) (sqr(daCeiling-ly));
 ysqrd_floor= (int) (sqr(ly-daFloor)  );

 	prof_enter(PROF_PREFRAME_MEMSET);
 memset(ptr, 0, xsize_render(xres())*ysize_render(yres())*4);
 	prof_leave(PROF_PREFRAME_MEMSET);
 // preframe things -----------------------------------
 	prof_enter(PROF_SPHERE_THINGS);
 btm = bTime();
 if (!(vframe&cd_frames)) { // first frame
 	for (i=0;i<spherecount;i++) {
		if (sp[i].flags & ANIMATED) {
			calculate_XYZ(sp+i, btm, obex);
			sp[i].x = obex[0];
			sp[i].y = obex[1];
			sp[i].z = obex[2];
			sp[i].xi = 0;
			sp[i].yi = 0;
			sp[i].zi = 0;
			sp[i].time = btm;

			}
		sp[i].time = btm;
		}
    camera_moved = 1;
 	}
 if (vframe%cd_frames) // every frame but the first
 for (i=0;i<spherecount;i++) {
 	if (sp[i].flags & ANIMATED) {
		calculate_XYZ(sp+i, btm, obex);
		sp[i].xi = (obex[0] - sp[i].x) / (btm - sp[i].time);
		sp[i].yi = (obex[1] - sp[i].y) / (btm - sp[i].time);
		sp[i].zi = (obex[2] - sp[i].z) / (btm - sp[i].time);
		camera_moved = 1;
		}
	}
	prof_leave(PROF_SPHERE_THINGS);
	prof_enter(PROF_PHYSICS);
 if (Physics) { // should we process physics?
 	camera_moved = 1;
 	for (i=0;i<spherecount;i++)
		apply_gravity(sp+i, btm); // first of all, Gravity
	for (i=0;i<spherecount;i++)
		apply_air(sp+i, btm, APPLY_ALL, 1);    // air resistance
	if (CollDetect)			 // do we process collisions?
	/* The algorythm is simple:
		We iterate thru all collideable spheres and check for collision.
		If a collision occurs, we update vectors (and other data), and set the collided balls'
		time to the moment of the collision. The balls are ready to collide again within the same
		time segment of the frame, but now they differ from the others, being advanced a little
		in the future (this is very important). After everything is completed, all balls' times
		are synced to the new time
	*/
		//first of all, we make a list with all the spheres, which "move"
		check_moving();
		for (j=0;j<spherecount;j++) if (sp[j].flags & COLLIDEABLE)
			for (i=j+1;i<spherecount;i++) if ((sp[i].flags & COLLIDEABLE) && can_collide(j, i))
				if ((ttm=collide(sp+i, sp+j, btm))>0) // collision detected
					ProcessIncident(sp+i, sp+j, ttm);
	/*
		Here is where advance takes place. All uncollided (and some collided, too)
		spheres are moved with respect to the remaining time to the sync time and
		their movement vectors, too.
	*/
	for (i=0;i<spherecount;i++)
		advance(sp+i, btm);
 	}
	prof_leave(PROF_PHYSICS);
	prof_enter(PROF_SORT);
 sort_fract_array(sp, spherecount);
 	prof_leave(PROF_SORT);

	prof_enter(PROF_PROJECT);
for (i=0;i<RES_MAXY;i++) {
	RowMin[i] = +65536;
	RowMax[i] = -65536;
}
for (i=0;i<spherecount;i++) {
 	projectIt(sp+i, pp[i], num_sides+i, ptr, cur, w, xsize_render(xres()), ysize_render(yres()), i+1);
	pdlo[i] = pith0(sp[i].x, sp[i].y, sp[i].z, lx, ly, lz);
 	sp[i].dist = pith0(sp[i].x, sp[i].y, sp[i].z, cur[0], cur[1], cur[2]);
	}
	prof_leave(PROF_PROJECT);
	prof_enter(PROF_PASS);
 PassPre(pdlo, pith(lw, cur), lw);
 	prof_leave(PROF_PASS);
}

/*
 *	Parameter initialisation routine - initialises the vector grid basics and interpolation increasers
 */
void bash_preframe(double *lw, double *tt, double *ti, double *tti)
{
 int i;
 lw[0] = lx; lw[1] = ly; lw[2] = lz;
 calc_grid_basics(cur, alpha, beta, w);
 for (i=0;i<3;i++) {
 	tt[i]=w[0][i];
	ti[i]=(w[1][i]-w[0][i])/((double) xsize_render(xres()));
	tti[i]=(w[2][i]-w[0][i])/((double) ysize_render(yres()));
 	}
}

#ifdef _MSC_VER
// returns the 2 base logarithm of x
double log2(double x)
{
	return log(x) * 1.4426950408889634073599246810019;
}
#endif
/*
 *	Chooses which texture instance to use based on the distance of the row from the camera and the angle
 *	the ray from the camera encloses with that row.
 */
void choose_texture(double *cur, int xr, int yr, double cx, double cxi, double cy, double cyi, double hx, double hy,
			double dp, int *xtex, int *xandmask, int *xshl, int *tam, int *notmask
			)
{
	double parea; // parea is the area (in square texels) which single ray covers
	int i;
	parea = sqrt(cxi * cxi + cyi * cyi) * sqrt(sqr(cx - hx) + sqr(cy - hy));
	*xtex = 0; *xandmask = MAX_TEXTURE_SIZE; *xshl = (int) log2(*xandmask);
	/*
		The idea is simple. Since we want to bilinear filter (or do nearest neighbour) all the time
		we do want a single ray to cover an area of approx 1 sq texel, we get the area a ray will have
		on the 256x256 texture. If the area is above `Area_const' (an meaningful value [1..4]) we want
		to get a smaller texture - a 128x128 version mipmap will have 4 times less texels, resulting in
		four times less area. So we get lower texture sizes and divide the area by 4 each time, until
		the area is acceptable or we reached using 1x1 texture.
	*/
	if (do_mipmap)
	while (parea>Area_const && *xtex<end_tex) {
		(*xtex)++; (*xandmask)/=2; (*xshl)--;
		parea /= 4.0;
		}
		else
	for (i=0;i<3;i++) {
		(*xtex)++; (*xandmask)/=2; (*xshl)--;
		parea /= 4.0;
	}
	(*xandmask)--;
	*notmask = (0xffffffff >> (LOG2_TEXTURE_AREA+*xtex)) << (LOG2_TEXTURE_AREA+*xtex);
	*tam = (1 << (2*(*xshl))) - 1;
}

/*************************************************
 * This is the non-SSE version of the renderer,  *
 * which is single-threaded and does all-in-one: *
 * pre-filling, sphere and floor rendering. Uses *
 * function calls, FPU and integer arithmetic    *
 *************************************************/
#ifdef ACTUALLYDISPLAY
void DrawIt_P5(SDL_Surface *p, SDL_Overlay *ov)
#else
void DrawIt_P5(void)
#endif
{int ii,i,j,k;
 int Lx=99999, Ly=99999, Lr=0, SavergX, SavergY;
 int Bx, By, notmask, Tc, boffset, texandmask;
 unsigned Fx, Fy;
 Uint32 *ptr;
 double t[3], tt[3], ti[3], tti[3], v[3], lw[3], ydist, scalefactor, dp, ut[3];
 double A, Arcp, pdet, pgB;

 int xtex, xandmask, xshl;

 float cx, cy, cxi, cyi, hx, hy;

 int gX, gY, gxi, gyi, xr, yr;

 Uint32 fcol=0, bfcol;
 float opacity,bopacity;
 int dropped, backd, foundsphere, addstuff;

#ifdef FSTAT
 good = worse = realbad = 0;
#endif

 xr = xsize_render(xres());
 yr = ysize_render(yres());

 bash_preframe(lw, tt, ti, tti);

#ifdef BLUR
 ptr = (Uint32 *)blurbuffer;
#else
 ptr = (Uint32 *) framebuffer;
#endif
 preframe_do(ptr, lw);

 for (j=0;j<yr;j++) {
 	for (k=0;k<3;k++) {
		t[k]=tt[k];
		ut[k] = tt[k] - tti[k];
	}
	// check for horizon
	if ((ydist=vfabs(cur[1]-t[1]))>DST_THRESHOLD) {
		dp = (cur[1]>t[1]?(cur[1]-daFloor):(daCeiling-cur[1]));
		ysqrd = (cur[1]>t[1]?(ysqrd_floor):(ysqrd_ceil));
		scalefactor = dp/ydist;
		cx = cur[0] + (t[0]-cur[0])*scalefactor;
		cy = cur[2] + (t[2]-cur[2])*scalefactor;
		cxi = (cur[0]+(t[0]+ti[0]*xr-cur[0])*scalefactor - cx)/((double) xr);
		cyi = (cur[2]+(t[2]+ti[2]*yr-cur[2])*scalefactor - cy)/((double) yr);
		// calculate second "helping" point from the previos image row
		scalefactor = (cur[1]>ut[1]?(cur[1]-daFloor):(daCeiling-cur[1])) / vfabs(cur[1]-ut[1]);
		hx = cur[0] + (ut[0]-cur[0])*scalefactor;
		hy = cur[2] + (ut[2]-cur[2])*scalefactor;

		gX = (int) (cx*65536.0);
		gY = (int) (cy*65536.0);
		gxi =(int) (cxi*65536.0);
		gyi =(int) (cyi*65536.0);
		// choose apropriate texture from the T array:
		choose_texture(cur, xr, yr, cx, cxi, cy, cyi, hx, hy, dp,
			&xtex, &xandmask, &xshl, &texandmask, &notmask);
		for (i=0;i<xr;i++)
			{
			addstuff = 0;
			dropped = *ptr;
			if (dropped) {
				backd = dropped>>16;
				dropped &= 0xffff;
				v[0] = t[0]-cur[0]; v[1] = t[1] - cur[1]; v[2] = t[2] - cur[2];
				A = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
				Arcp = 1.0/A;
				//norm(v); normalize it later... :)
				if (testsphere1(sp+dropped-1, v, cur, A, &pdet, &pgB)) {
					fcol = Solve3D(sp + dropped-1 , v, cur, lw, Arcp, &opacity, &pdet, &pgB, 0);
					if (opacity > 0.99)  *ptr=fcol;
					               else {
						       	if (backd) {
						       		if (testsphere1(sp+backd-1, v, cur, A, &pdet, &pgB)) {
									bfcol = Solve3D(sp + backd - 1, v, cur, lw, Arcp,
									                &bopacity, &pdet, &pgB, 0);
									if (bopacity>0.99) *ptr = blend_p5(fcol, bfcol, opacity);
										else	 {
										fcol = blend_p5(fcol, bfcol, opacity);
										opacity = opacity + (bopacity*(1-opacity));
										addstuff=1;
										dropped=0;
										}
						       			}
									else {*ptr=fcol; addstuff=1; dropped = 0;}
								}
							else{addstuff = 1; dropped = 0;}
						       }
#ifdef FSTAT
					good++;
#endif
					}
					else {// too bad - we'll have to iterate all the spheres to find the right one...
					foundsphere = 0;
#ifdef FSTAT
					worse++;
#endif
					for (ii=spherecount-1;ii>=0;ii--)
						if (testsphere1(sp + ii, v, cur, A, &pdet, &pgB)) { // sphere found...
						fcol = Solve3D(sp + ii, v, cur, lw, Arcp, &opacity, &pdet, &pgB, 0);
						if (opacity > 0.99) *ptr=fcol;
						               else {addstuff = 1; dropped = 0; }
						foundsphere = 1;
						break;
						}
					if (!foundsphere)
						{
						/******************************************************
						 * If we reached here, this means that the pre-filler *
						 * misled us there is a sphere... we've done alot a   *
						 * work to make it here, so the only thing that keeps *
						 * our engine from renderering *REALLY* slow is that  *
						 * such BAD pixels are very rare (about 1-2 %)        *
						 ******************************************************/
#ifdef FSTAT
						 realbad++;
#endif
						dropped = 0; addstuff=0;
						}
					}
				}
			if (!dropped) {
				if (!bilfilter) {
					SavergX = r(gX); SavergY = r(gY);
					if (SavergX==Lx && SavergY==Ly)
						*(ptr) = Lr;
						else{
						Lx=SavergX; Ly=SavergY;
						*(ptr) = Lr = multiplycolor(T[xtex].data[(((unsigned) (Lx)>>xtex) & xandmask) + (((((unsigned)(Ly)>>xtex) & xandmask))<<xshl)],lform(Lx, Ly, lx, lz, ysqrd));
						}
					}
					else {
					Bx = (gX<0?(notmask|(gX>>(16+xtex))):(gX>>(16+xtex)));
					By = (gY<0?(notmask|(gY>>(16+xtex))):(gY>>(16+xtex)));
					Fx = ((gX>>xtex)&0xffff);
					Fy = ((gY>>xtex)&0xffff);
					boffset = ((Bx & xandmask)) + (((By & xandmask))<<xshl);
					Tc = bilinea_p5(	T[xtex].data[boffset				],
								T[xtex].data[(boffset+1) &texandmask		],
								T[xtex].data[(boffset+xandmask+1)&texandmask	],
								T[xtex].data[(boffset+xandmask+2)&texandmask	],
								Fx, Fy);
					*(ptr) = multiplycolor(Tc, lform(r(gX), r(gY), lx, lz, ysqrd));
					}
				}
			if (addstuff) *ptr = blend_p5(fcol, *ptr, opacity);
			// continue interpolation
			ptr++;
			gX += gxi; gY+=gyi;
			t[0]+=ti[0]; t[2]+=ti[2];
			}
		}
		else {
 		for (i=0;i<xr;i++)
			{
			addstuff = 0;
			dropped = (*ptr);
			if (dropped) {
				backd = dropped>>16;
				dropped &= 0xffff;
				v[0] = t[0]-cur[0]; v[1] = t[1] - cur[1]; v[2] = t[2] - cur[2];
				A = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
				Arcp = 1.0/A;
				//norm(v);
				if (testsphere1(sp+dropped-1, v, cur, A, &pdet, &pgB)) {
					fcol = Solve3D(sp + dropped-1 , v, cur, lw, Arcp, &opacity, &pdet, &pgB, 0);
					if (opacity > 0.99)  *ptr=fcol;
						else{
						       	if (backd) {
						       		if (testsphere1(sp+backd-1, v, cur, A, &pdet, &pgB)) {
									bfcol = Solve3D(sp + backd - 1, v, cur, lw, Arcp,
									                &bopacity, &pdet, &pgB, 0);
									if (bopacity>0.99) *ptr = blend_p5(fcol, bfcol, opacity);
										else {
										fcol = blend_p5(fcol, bfcol, opacity);
										opacity = opacity + (bopacity*(1-opacity));
										addstuff=1;
										dropped=0;
										}
						       			}
									else {*ptr=fcol; addstuff=1; dropped = 0;}
								}
							else{addstuff = 1; dropped = 0;}
						       }
#ifdef FSTAT
					good++;
#endif
					}
					else {// too bad - we'll have to iterate all the spheres to find the right one...
#ifdef FSTAT
					worse++;
#endif
					foundsphere = 0;
					for (ii=spherecount-1;ii>=0;ii--)
						if (testsphere1(sp + ii, v, cur, A, &pdet, &pgB)) { // sphere found...
						fcol = Solve3D(sp + ii, v, cur, lw, Arcp, &opacity, &pdet, &pgB, 0);
						if (opacity > 0.99) *ptr=fcol;
						               else {addstuff = 1; dropped = 0; }
						foundsphere = 1;
						break;
						}
					if (!foundsphere)
						{
#ifdef FSTAT
						 realbad++;
#endif
						 dropped = 0; addstuff=0;
						}
					}
				}
			if (addstuff) *ptr = multiplycolorf(fcol, opacity);
			ptr++;
			gX += gxi; gY+=gyi;
			t[0]+=ti[0]; t[2]+=ti[2];
			}
		}
	for (k=0;k<3;k++) tt[k]+=tti[k];
 	}
// frame generation finished:
#ifdef ACTUALLYDISPLAY
 postframe_do(p, ov, framebuffer);
#else
 postframe_do(framebuffer);
#endif
}

/*
	render_floor:

	This is a part of the SSE version of the renderer. This creates an image of the floor (and the ceiling)
	only. Shadow-casting is not performed here. Uses assembly-optimized code, with loop unrolling and not
	a single IF in the whole routine :)))

 */

#ifdef USE_ASSEMBLY
SSE_ALIGN(static const float c64k[4]) =
	{BRIGHTNESS_C, BRIGHTNESS_C, BRIGHTNESS_C, BRIGHTNESS_C};
SSE_ALIGN(static const unsigned short my_ffs[8]) =
	{0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x0, 0xffff, 0x0};

#endif

void render_floor(Uint32 *dptr, int xr, int yr, double *tt, double *ti, double *tti, int start_line)
{
	int i, j, k;
	float fx[4], fy[4];
	float fxi[4];
	float fyi[4];
	double ydist, dp, scalefactor, cx, cy, cxi, cyi, ut[3], hx, hy;
	int xtex, xandmask, xshl, texandmask, notmask;
	int xxandmask[2];
	Uint32 *ptex;

	// these alignments are needed to use movaps intead of movups.
	SSE_ALIGN(float fxyi[8]);
	SSE_ALIGN(float lightxy1[16]);

#ifdef USE_ASSEMBLY
	SSE_ALIGN(int save[4]);
#endif
	SSE_ALIGN(int gx[4]);
	SSE_ALIGN(int gy[4]);
	SSE_ALIGN(int gxx[8]);
	//printf("thread #%d ran render_floor(); addr. of lightxy1 = 0x%X\n", start_line, (unsigned) lightxy1);	fflush(stdout);
	for (k=0;k<3;k++) tt[k] += start_line * tti[k];
	dptr += start_line * xr;
	for (j=start_line;j<yr;j+=cpu_count) {
		for (k=0;k<3;k++) ut[k] = tt[k] - tti[k];
		if ((ydist=vfabs(cur[1]-tt[1]))>DST_THRESHOLD) {
				prof_enter(PROF_INIT_PER_ROW);
			dp = (cur[1]>tt[1]?(cur[1]-daFloor):(daCeiling-cur[1]));

			// load y distance, squared
			lightxy1[15] = lightxy1[14] = lightxy1[13] = lightxy1[12] = (cur[1]>tt[1]?ysqrd_floor:ysqrd_ceil);

			scalefactor = dp/ydist;
			cx = cur[0] + (tt[0]-cur[0])*scalefactor;
			cy = cur[2] + (tt[2]-cur[2])*scalefactor;
			cxi = (cur[0]+(tt[0]+ti[0]*xr-cur[0])*scalefactor - cx)/((double) xr);
			cyi = (cur[2]+(tt[2]+ti[2]*yr-cur[2])*scalefactor - cy)/((double) yr);
			// calculate second "helping" point from the previos image row
			scalefactor = (cur[1]>ut[1]?(cur[1]-daFloor):(daCeiling-cur[1])) / vfabs(cur[1]-ut[1]);
			hx = cur[0] + (ut[0]-cur[0])*scalefactor;
			hy = cur[2] + (ut[2]-cur[2])*scalefactor;

			for (i=0;i<4;i++) {
				gxx[i] = (int) (cxi*262144.0);
				gxx[i+4] = (int) (cyi*262144.0);
				gx[i] = (int) (cx*65536.0) + (int) (i*cxi*65536.0);
				gy[i] = (int) (cy*65536.0) + (int) (i*cyi*65536.0);
				fx[i] = cx + i*cxi;
				fy[i] = cy + i*cyi;
				fxi[i] = cxi*4;
				fyi[i] = cyi*4;
				lightxy1[i] = lx;
				lightxy1[i+4] = lz;
				lightxy1[i+8] = 1.0;
				fxyi[  i] = fxi[i];
				fxyi[4+i] = fyi[i];
			}
			// choose apropriate texture from the T array:
			choose_texture(cur, xr, yr, cx, cxi, cy, cyi, hx, hy, dp,
				&xtex, &xandmask, &xshl, &texandmask, &notmask);
			xxandmask[0] = xxandmask[1] = xandmask;
			// actual rendering follows:
			ptex = (Uint32 *) T[xtex].data;
			i = xr/4;
				prof_leave(PROF_INIT_PER_ROW);
				prof_enter(PROF_WORK_PER_ROW);
			// load some useful values:
#define render1
#include "x86_asm.h"
#undef render1
			if (bilfilter) {
				while (i--) {
#define render2
#include "x86_asm.h"
#undef render2
					dptr+=4;
				}
			} else {
			// NEAREST NEIGHBOUR FILTER VERSION:
				while (i--) {
#define render3
#include "x86_asm.h"
#undef render3
					dptr += 4;
				}
			}
			prof_leave(PROF_WORK_PER_ROW);
		} else { // fill the row with zeros:
			memset(dptr, 0, 4*xr);
			dptr+=xr;
		}
#define emit_emms
#include "x86_asm.h"
#undef emit_emms
		for (i=0;i<3;i++) tt[i]+=cpu_count*tti[i];
		dptr += (cpu_count-1)*xr;
	}
}

void render_spheres_init(unsigned short *fbuffer)
{
	int xr = xsize_render(xres()), yr = ysize_render(yres());
	prof_enter(PROF_MEMSETS);
	memset(fbuffer, 0, xr*yr*sizeof(unsigned short));
	memset(RowInfo, 0, sizeof(RowInfo));
	prof_leave(PROF_MEMSETS);
}

// this is for the SSE version of the renderer, too. Renders the spheres in fbuffer and puts some
// blending data in fbuffer (zeros in fbuffers means there is no sphere for the specified pixel; see
// merge_rows in fract_asm.h for details)

// NOTE: this uses a modified code from DrawIt_P5
void render_spheres(Uint32 *fb, unsigned short *fbuffer,
		double *tt, double *ti, double *tti, int start_line)
{
	int i, j, k, xr, yr, dropped, backd, ii, x_start, x_end;
	double t[3];
	double v[3];
	double lw[3];
	double A, Arcp, pdet, pgB;
	float opacity, bopacity;
	Uint32 fcol, bfcol;
	int foundsphere;
	xr = xsize_render(xres()); yr = ysize_render(yres());
	lw[0] = lx; lw[1] = ly; lw[2] = lz;

	//printf("Thread #%d ran render spheres\n", start_line); fflush(stdout);

	for (k=0;k<3;k++) tt[k] += start_line * tti[k];
	fb += start_line * xr;
	fbuffer += start_line * xr;

	for (j=start_line;j<yr;j+=cpu_count) {
		if (RowMin[j] >= xr || RowMax[j] < 0) { // skip whole row
			RowInfo[j] = 0;
			fbuffer += xr;
			fb += xr;
		} else {
			x_start = RowMin[j] < 0 ? 0 : RowMin[j];
			x_end   = RowMax[j] >= xr ? (xr-1) : RowMax[j];
			for (k=0;k<3;k++) t[k] = tt[k] + ti[k] * x_start;
			fbuffer += x_start;
			fb += x_start;
			for (i=x_start;i<=x_end;i++,fb++,fbuffer++,t[0]+=ti[0],t[2]+=ti[2]) {
				if (*fb) {
					dropped = *fb;
					backd = dropped >> 16;
					dropped &= 0xffff;
					//if (dropped > spherecount) {printf("***bad dropped(%d);***\n", dropped);fflush(stdout);return;}
					v[0] = t[0]-cur[0]; v[1] = t[1] - cur[1]; v[2] = t[2] - cur[2];
					//norm(v);
					A = v[0]*v[0] + v[1]*v[1] + v[2]*v[2];
					Arcp = 1.0/A;

					if (testsphere1(sp+dropped-1, v, cur, A, &pdet, &pgB)) {
						fcol = Solve3D(sp + dropped-1 , v, cur, lw, Arcp, &opacity, &pdet, &pgB, 0);
						if (opacity > 0.99)  {
							*fb=fcol;
							*fbuffer = 65535;
						} else {
							if (backd) {
								if (testsphere1(sp+backd-1, v, cur, A, &pdet, &pgB)) {
									bfcol = Solve3D(sp + backd-1, v, cur, lw, Arcp,
								                	&bopacity, &pdet, &pgB, 0);
									if (bopacity>0.99) {
										*fb = blend_sse(fcol, bfcol, opacity);
										*fbuffer = 65535;
									} else {
										*fb = blend_sse(fcol, bfcol, opacity);
										opacity = opacity + (bopacity*(1-opacity));
										*fbuffer = (unsigned int) (opacity*65535);
									}
								} else {
								*fb=fcol;
									*fbuffer = (unsigned int) (opacity*65535);
								}
							} else {
								*fb = fcol;
								*fbuffer = (unsigned int) (opacity*65535);
							}
						}
					} else {// too bad - we'll have to iterate all the spheres to find the right one...
						foundsphere = 0;
						for (ii=spherecount-1;ii>=0;ii--)
							if (testsphere1(sp + ii, v, cur, A, &pdet, &pgB)) { // sphere found...
								fcol = Solve3D(sp + ii, v, cur, lw, Arcp, &opacity, &pdet, &pgB, 0);
								if (opacity > 0.99) {
									*fb=fcol;
									*fbuffer = 65535;
								} else {
									*fb = fcol;
									*fbuffer = (unsigned int) (opacity*65535);
								}
								foundsphere = 1;
								break;
							}
					}
					if (*fbuffer) RowInfo[j] = 1;
				}
			}
			fb += (xr-1-x_end);
			fbuffer += (xr-1-x_end);
		}
		for (k=0;k<3;k++) tt[k] += cpu_count * tti[k];
		fb += (cpu_count-1) * xr;
		fbuffer += (cpu_count-1) * xr;
	}
}

static float rowstart[RES_MAXY][2], rowincrease[RES_MAXY][2];
static int   queue[SHADOW_QUEUE_SIZE][2];
const int xi[4] = {+1, -1,  0,  0};
const int yi[4] = { 0,  0, +1, -1};
static int   transp_multi[MAX_SPHERES]; // storage for each sphere's shadow intensity
static float thresh_D[MAX_SPHERES]; // Discriminant threshold for each sphere. If D > thresh, full intensity shadow is applied
static float rthresh_D[MAX_SPHERES];
int ceilmax, floormin; // ending row for ceiling, starting row for floor in the image

void shadow_fill(unsigned short *sbuffer, int xr, int yr, int cx, int cy, int su)
{
	int i, j, k, nx, ny, u, v, oof;
	unsigned short current_sp, n_pixel, old_pixel;
	float a, b, c, xp, yp, zp, D;

	current_sp = (1 + (su%S_SPNUM_MODULUS)) << S_SPNUM_OFFSET;
	xp = lx - sp[su].x;
	yp = ly - sp[su].y;
	zp = lz - sp[su].z;
	i = 0;
	j = 2;
	queue[0][0] = cx; queue[0][1] = cy;
	queue[1][0] = queue[1][1] = QSENTINEL;
	while (j-i != 1) {
		if (queue[i][0] != QSENTINEL) {
			// check whether it is shadow point.
			u = queue[i][0]; v = queue[i][1];
			a = rowstart[v][0] + rowincrease[v][0]*u - lx;
			b = (v>=floormin?daFloor:daCeiling) - ly;
			c = rowstart[v][1] + rowincrease[v][1]*u - lz;
			D = sqr(xp*a+yp*b+zp*c) - (a*a+b*b+c*c)*(xp*xp + yp*yp + zp*zp - sqr(sp[su].d));
			if (D >= 0.0f) {
				D /= (a*a+b*b+c*c);
				if (D > thresh_D[su])
					n_pixel = transp_multi[su];
					else
					n_pixel = (int) (transp_multi[su]*D*rthresh_D[su]);
				oof = u + xr*v;
				if (sbuffer[oof] & S_MIXED_MASK) {
					old_pixel = S_MAX_BLACK - (sbuffer[oof] & S_COLOR_MASK);
					sbuffer[oof] = S_MIXED_MASK | current_sp |
						(S_MAX_BLACK - ((old_pixel*(S_MAX_BLACK-n_pixel))>>S_COLOR_BITS));
				} else {
					sbuffer[oof] = n_pixel | current_sp;
				}
				for (k=0;k<4;k++) {
					nx = u + xi[k];
					ny = v + yi[k];
					oof = ny*xr + nx;
					if (((unsigned) nx)<(unsigned)xr &&
					    ((unsigned) ny)<(unsigned)yr &&
					    (!sbuffer[oof] || (!(sbuffer[oof]&S_MIXED_MASK)&&
					    ((sbuffer[oof]&S_SPNUM_MASK)!=current_sp)/*&&
					    (sbuffer[ny*xr+nx]&S_COLOR_MASK)!=S_MAX_BLACK*/))
					    ) {
						if (sbuffer[oof]) // prevent from multiple inclusion
							sbuffer[oof] |= S_MIXED_MASK;
							else
							sbuffer[oof] = current_sp;

						queue[j][0] = nx;
						queue[j++][1] = ny;
						j&=qmod;
						if (j==i) {
							return;
						}
				    	}
				}
			}
		} else {
			queue[j++][0] = QSENTINEL;
			j &= qmod;
		}
		i = (i+1) & qmod;
	}
}

static inline int check_sphere_shadow(int x, int y)
{
	int i;
	float xp, yp, zp, a, b, c;
	for (i=0; i < spherecount; i++) {
		xp = lx - sp[i].x;
		yp = ly - sp[i].y;
		zp = lz - sp[i].z;
		a = rowstart[y][0] + rowincrease[y][0]*x - lx;
		b = (y>=floormin?daFloor:daCeiling) - ly;
		c = rowstart[y][1] + rowincrease[y][1]*x - lz;

		if (xp*a+yp*b+zp*c < 0.0f &&
			sqr(xp*a+yp*b+zp*c) - (a*a+b*b+c*c)*(xp*xp + yp*yp + zp*zp - sqr(sp[i].d))>0.0f) return i;
	}
	return -1;
}

void check_generic(unsigned short *sbuffer, int xr, int yr, int x, int y)
{
	int su;
	if (x < 0 || x >= xr || y < 0 || y >= yr) return;
	if (y > ceilmax && y < floormin) return;
	if (!sbuffer[x + y*xr]) {
		su = check_sphere_shadow(x, y);
		if (su >= 0)
			shadow_fill(sbuffer, xr, yr, x, y, su);
	}
}

void render_shadows(Uint32 *floorbuffer, unsigned short *sbuffer, int xr, int yr, double *mtt, double *mti, double *mtti)
{
	int i, j, su, u, v, ckx, cky;
	double planeDist, y, m, ml[3] = {lx, ly, lz};


	ceilmax = floormin = -1;
	prof_enter(PROF_ZERO_SBUFFER); // reset the S-buffer;
	memset(sbuffer, 0, xr*yr*sizeof(unsigned short));
	prof_leave(PROF_ZERO_SBUFFER);

	prof_enter(PROF_SHADOW_PRECALC);
	for (j = 0; j < yr; j++) {
		y = mtt[1] + j*mtti[1] - cur[1];
		/*if (fabs(y) < DST_THRESHOLD) {
			if (ceilmax==-1) ceilmax = j - 1;
			continue;
		}*/
		if (y > 0.0 && fabs(y) > DST_THRESHOLD) ceilmax = j;
		if (floormin == -1 && y < 0.0) floormin = j;
		planeDist = (y<0.0)?(cur[1] - daFloor):(daCeiling - cur[1]);
		m = planeDist / fabs(y);
		for (i=0;i<2;i++) {
			rowstart   [j][i] = cur[i*2] + (mtt[i*2]+mtti[i*2]*j-cur[i*2])*m;
			rowincrease[j][i] = mti[i*2] * m;
		}
	}
	if (floormin == -1) floormin = yr;
	for (i = 0; i < spherecount; i++) {
		if (sp[i].flags & TRANSPARENT)
			transp_multi[i] = (int) ((S_DEFAULT_OPACITY+(1.0-S_DEFAULT_OPACITY)*(1.0-sp[i].opacity)) * S_MAX_BLACK);
			else
			transp_multi[i] = S_MAX_BLACK;
		thresh_D[i] = S_EDGE * sp[i].d * sp[i].d;
		rthresh_D[i] = 1.0f/thresh_D[i];
	}
	prof_leave(PROF_SHADOW_PRECALC);

	prof_enter(PROF_SHADOWIZE);
	for (su=0;su<spherecount;su++) {
		if (ProjectSphereShadow(sp + su, ml, &u, &v, xr, yr, cur, mtt, mti, mtti)) {
			//if ((sbuffer[u + xr*v]&0x0f00) != 0) continue;
			shadow_fill(sbuffer, xr, yr, u, v, su);
		}
	}
	prof_leave(PROF_SHADOWIZE);

	prof_enter(PROF_CANDIDATESEARCH);
	//check for spheres not rendered due to center being outside the screen:
	// check first row:
	for (ckx = 0; ckx < xr; ckx+=CANDIDATE_INTERVAL) {
		check_generic(sbuffer, xr, yr, ckx, 0);
		check_generic(sbuffer, xr, yr, ckx, yr-1);
		check_generic(sbuffer, xr, yr, ckx, ceilmax);
		check_generic(sbuffer, xr, yr, ckx, floormin);
	}
	for (cky = 0; cky < yr; cky+=CANDIDATE_INTERVAL) {
		check_generic(sbuffer, xr, yr, 0, cky);
		check_generic(sbuffer, xr, yr, xr-1, cky);
	}
	prof_leave(PROF_CANDIDATESEARCH);

	prof_enter(PROF_MERGE);
	//for (i=0;i<xr*yr;i++)
	//	floorbuffer[i] = multiplycolor(floorbuffer[i], 255 * (255 - (255&sbuffer[i])));
	shadows_merge_mmx2(floorbuffer, sbuffer, xr*yr);
	prof_leave(PROF_MERGE);
}

void merge_buffers(Uint32 *src, Uint32 *dst, unsigned short *fbuffer)
{
	int y, xr, yr;
	xr = xsize_render(xres()); yr = ysize_render(yres());
	for (y = 0; y < yr; y++, src+=xr, dst+=xr, fbuffer+=xr) if (RowInfo[y]) {
		merge_rows(src, dst, fbuffer, xr);
	}
	//memcpy(sph, flr, xsize_render(xres()) * ysize_render(yres()) * 4);
}

/*
	This is the SSE version of the renderer (well, this routine is merely a nasty caller)
	The idea is:

	1. Render the floor very fast, using vectorized SSE code in render_floor;
	2. Render the spheres using less innermost cycle IFs, and preserving alpha data in a seperate buffer;
	3. If we render shadows, calculate them in seperate shadow buffer and merge them to the floorbuffer
	3. Blend the two framebuffers into a single with merge_buffers (using MMX);
	4. For everything else use the shared code for the P5 and SSE versions (bash_preframe,
		preframe_do, postframe_do)

	The good news is, that, floor rendering is up to 5 times faster;
	The bad news are, that sphere rendering is a bit slower, and there is an additional overhead (~2%) for
	framebuffer merge. Also, pixels that are rendered with the floor renderer behind the balls are
	calculated in vain, 'coz they are lost forever.
*/
#ifdef ACTUALLYDISPLAY
void DrawIt_SSE(SDL_Surface *p, SDL_Overlay *ov)
#else
void DrawIt_SSE(void)
#endif
{
	double lw[3], tt[3], ti[3], tti[3];
	double t0[3];
	int xr, yr, i;
	Uint32 *ptr;

	xr = xsize_render(xres()); yr = ysize_render(yres());

#ifdef BLUR
	ptr = blurbuffer;
#else
	ptr = floorbuffer;
#endif
	prof_enter(PROF_BASH_PREFRAME);		bash_preframe(lw, tt, ti, tti);			prof_leave(PROF_BASH_PREFRAME);
	prof_enter(PROF_PREFRAME_DO);		preframe_do( framebuffer, lw);			prof_leave(PROF_PREFRAME_DO);

	//save tt's
	for (i=0;i<3;i++) { t0[i] = tt[i];}

	if (cpu_count==1) {
	prof_enter(PROF_RENDER_FLOOR);
		render_floor(ptr, xr, yr, tt, ti, tti, 0);
	prof_leave(PROF_RENDER_FLOOR);

	for (i = 0; i < 3; i++) tt[i] = t0[i]; //restore tt's

	prof_enter(PROF_RENDER_SPHERES);
		render_spheres_init(fbuffer);
		render_spheres(framebuffer, fbuffer, tt, ti, tti, 0);
	prof_leave(PROF_RENDER_SPHERES);

	} else {
		thread_main(t0, ti, tti, xr, yr, ptr);
	}
	if (r_shadows){
	prof_enter(PROF_RENDER_SHADOWS);	render_shadows(ptr, sbuffer, xr, yr, t0, ti, tti);	prof_leave(PROF_RENDER_SHADOWS);
	}
	prof_enter(PROF_MERGE_BUFFERS);		merge_buffers(ptr, framebuffer, fbuffer);		prof_leave(PROF_MERGE_BUFFERS);
#ifdef ACTUALLYDISPLAY
	prof_enter(PROF_POSTFRAME_DO);		postframe_do(p, ov, floorbuffer);			prof_leave(PROF_POSTFRAME_DO);
#else
	prof_enter(PROF_POSTFRAME_DO);		postframe_do(floorbuffer);				prof_leave(PROF_POSTFRAME_DO);
#endif
}

/****************************************************************
 * This is the master function which chooses who to call to do  *
 * the actual job. Currently, the "CPU world" is divided to     *
 * SSE and non-SSE only.                                        *
 ****************************************************************/
#ifdef ACTUALLYDISPLAY
void render_single_frame(SDL_Surface *p, SDL_Overlay *ov)
{
	if ( sse_enabled )
		DrawIt_SSE(p, ov);
		else
		DrawIt_P5(p, ov);
}

#else
void render_single_frame(void)
{
	printf("."); fflush(stdout);
	if ( sse_enabled )
		DrawIt_SSE();
		else
		DrawIt_P5();
}
#endif


void set_cpus(int new_count)
{
	if (new_count < 1) {
		printf("Invalid processor count: %d\n", new_count);
		new_count = 1;
	} else if (new_count > MAX_CPU_COUNT) {
		printf("Sorry. Maximum allowable CPU count is %d\n", MAX_CPU_COUNT);
		new_count = MAX_CPU_COUNT;
	}
	cpu_count = new_count;
}

#ifdef ACTUALLYDISPLAY
SDL_Thread* r_threads[MAX_CPU_COUNT];
SDL_sem* r_sem[MAX_CPU_COUNT];
SDL_sem *main_sem, *start_sem;
#else
pthread_t r_threads[MAX_CPU_COUNT];
sem_t r_sem[MAX_CPU_COUNT];
sem_t main_sem, start_sem;
#endif
int threads_first = 1;
volatile int cmd_thread;
volatile int trd_counter;
static double thread_tt[3], thread_ti[3], thread_tti[3];
static Uint32 *thread_ptr;
static int thread_xres, thread_yres;

#ifdef ACTUALLYDISPLAY
int    fract_thread_dummy(void *data);
#else
void * fract_thread_dummy(void *data);
#endif

void thread_init(void)
{
	int i, res;
	int tstorage[MAX_CPU_COUNT];
	trd_counter = 0;
	for (i=0;i<cpu_count;i++) {
#ifdef ACTUALLYDISPLAY
		r_sem[i] = SDL_CreateSemaphore(0); // create n LOCKED semaphores
		res = (r_sem[i] == NULL);
#else
		res = sem_init(r_sem + i, 0, 0); // create n LOCKED semaphores
#endif
		if (res) printf("error creating %d-th semaphore\n", i);
	}
#ifdef ACTUALLYDISPLAY
	main_sem = SDL_CreateSemaphore(0); // create a semaphore for the main thread
	res = (main_sem == NULL);
#else
	res = sem_init(&main_sem, 0, 0);  // create a semaphore for the main thread
#endif
	if (res) printf("Unable to create the main semaphore\n");
#ifdef ACTUALLYDISPLAY
	start_sem = SDL_CreateSemaphore(0);  // create a semaphore for the rendering union
	res = (start_sem == NULL);
#else
	res = sem_init(&start_sem, 0, 0);    // create a semaphore for the rendering union
#endif
	if (res) printf("Unable to create the start semaphore\n");

	for (i=0;i<cpu_count;i++) {
		tstorage[i] = i;
#ifdef ACTUALLYDISPLAY
		r_threads[i] = SDL_CreateThread(fract_thread_dummy, &tstorage[i]); // create and run n threads
		res = (r_threads[i] == NULL);
#else
		res = pthread_create(r_threads + i, NULL, fract_thread_dummy, &tstorage[i]);
#endif
		if (res) printf("Error creating %d-th thread\n", i);
	}
	cmd_thread = THREAD_RUN_CMD;
}


#ifdef __GNUC__
void __attribute__((noinline,stdcall)) fract_thread(int data)
#else
void fract_thread(int data)
#endif
{
//	int trd_n = data;
	int trd_n = trd_counter++; // <- WARNING: THIS SHOULD BE DANGEROUS IF IT'S NOT ATOMIC!!!!!
	int i, xr, yr;
	double t0[3], tt[3], ti[3], tti[3];
	Uint32 *ptr;
	if (trd_n < 0 || trd_n > cpu_count) {
		printf("Thread is unexpectedly corrupted, disable multithreading for sanity!\n");
		exit(1) ;
	}
	// all threads become blocked after creation in the line below.
#ifdef ACTUALLYDISPLAY
	SDL_SemWait(start_sem); // initial wait
#else
	sem_wait(&start_sem);
#endif
	// they are unblocked when first frame is ready to be rendered
	while (cmd_thread != THREAD_QUIT_CMD) {
		// do stuff...
		for (i=0;i<3;i++) {
			tt[i] = thread_tt[i];
			ti[i] = thread_ti[i];
			tti[i]= thread_tti[i];
			t0[i] = thread_tt[i];
		}
		xr = thread_xres;
		yr = thread_yres;
		ptr = thread_ptr;
		render_floor(ptr, xr, yr, tt, ti, tti, trd_n);
		render_spheres(framebuffer, fbuffer, t0, ti, tti, trd_n);

#ifdef ACTUALLYDISPLAY
		SDL_SemPost(main_sem);
#else
		sem_post(&main_sem);
#endif
	//	printf("t%d: waiting for my sem\n", trd_n); fflush(stdout);
#ifdef ACTUALLYDISPLAY
		SDL_SemWait(r_sem[trd_n]);
#else
		sem_wait(r_sem + trd_n);
#endif
	}
}

#ifdef ACTUALLYDISPLAY
#ifdef __GNUC__
int __attribute__((noinline)) fract_thread_dummy(void *data)
#else
int fract_thread_dummy(void *data)
#endif
#else
void* __attribute__((noinline)) fract_thread_dummy(void *data)
#endif
{
	register int thread_id_ = *((int*)data);
#ifdef __GNUC__
	__asm __volatile ("andl $-16,%%esp" ::: "%esp");
#endif
	fract_thread(thread_id_);
	return 0;
}


void thread_main(double *t0, double *ti, double *tti, int xr, int yr, Uint32 *ptr)
{
	int i;
	for (i=0;i<3;i++) { // pass the needed arguments to the threads:
		thread_tt[i] = t0[i];
		thread_ti[i] = ti[i];
		thread_tti[i]= tti[i];
	}
	thread_xres = xr;
	thread_yres = yr;
	thread_ptr = ptr;
	render_spheres_init(fbuffer); // clear the buffer before any thread has begun touching it ...

	// run the threads
	for (i=0;i<cpu_count;i++) {

		if (threads_first) {
#ifdef ACTUALLYDISPLAY
			SDL_SemPost(start_sem);
#else
			sem_post(&start_sem);
#endif
		}else {
#ifdef ACTUALLYDISPLAY
			SDL_SemPost(r_sem[i]);
#else
			sem_post(r_sem+i);
#endif
		}
	}

	/* right not the threads are running like idiots :)
	*/
	threads_first = 0;
	for (i=0;i<cpu_count;i++) {
#ifdef ACTUALLYDISPLAY
		SDL_SemWait(main_sem);
#else
		sem_wait(&main_sem);
#endif
	}
	// Once all threads are ready they all become blocked in the end of their while(...) { ...} cycle
}

void thread_finalize(void)
{
	int i;
	cmd_thread = THREAD_QUIT_CMD; // tell 'em to quit
#ifdef ACTUALLYDISPLAY
	SDL_Delay(35);
#endif
	for (i=0;i<cpu_count;i++) // unleash the power :)
#ifdef ACTUALLYDISPLAY
		SDL_SemPost(r_sem[i]);
#else
		sem_post(r_sem + i );
#endif
	for (i=0;i<cpu_count;i++) {
#ifdef ACTUALLYDISPLAY
		SDL_WaitThread(r_threads[i], NULL); // wait for completion (=free resources)
#else
		pthread_join(r_threads[i], NULL);
#endif
	}

#ifdef ACTUALLYDISPLAY
	for (i=0;i<cpu_count;i++)
		SDL_DestroySemaphore(r_sem[i]); // destroy thread semaphores
	SDL_DestroySemaphore(main_sem); // destroy main semaphore
	SDL_DestroySemaphore(start_sem);
#else
	for (i=0;i<cpu_count;i++)
		sem_destroy(r_sem+i);
	sem_destroy(&main_sem);
	sem_destroy(&start_sem);

#endif
}



void render_close(void)
{
	thread_finalize();
}
