/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *   rgb2yuv.cpp - handles RGB-to-YUV colorspace convesions.               *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "MyTypes.h"
#include "bitmap.h"
#include "cmdline.h"
#include "fract.h"
#include "rgb2yuv.h"


CPUCaps CPU;
int bestmethod=0;

Uint32 checksums[2][2][4] = {
	{ /*AMD*/
/*LinuX*/	{ 0xF781A462, 0xF781A462, 0xE06D540B, 0xFE9A01EB },
/*WindowS*/	{ 0xEBFD23B6, 0xEBFD23B6, 0xD4E8D4D0, 0x03A4FE3F },
	},
	{ /*Intel*/
/*LinuX*/	{ 0xF781A462, 0xF781A462, 0xF781A462, 0xF781A462 },
/*WindowS*/	{ 0xEBFD23B6, 0xEBFD23B6, 0xD4E8D4D0, 0xF33BFC69 }
	}
};

Uint32 csc=0xffffffff;

// these are used for benchmarking purposes
Uint32 fakeRGBBuffer[BENCHSIZE], fakeYUVBuffer[BENCHSIZE];

// the 64 MB bigarray...
Uint32 *bigarray = NULL;

// ConvertRGB2YUV_X86 - convers [count] pixels from [*src] to [*dest]
// transformation is FROM RGB colorspace TO YUY2 packed format
// the only portable function  :) Uses i386 arighmetics only. Fairly precise
// count must be even
void ConvertRGB2YUV_X86(Uint32 *dest, Uint32 *src, size_t count)
{register int r, g, b;
 int r1, g1, b1;
 count /= 2;
 while (count--) {
 	b = (*src) & 0xff;
	g = ((*src)>>8) & 0xff;
	r = ((*src)>>16) & 0xff;
	src++;
 	b1 = (*src) & 0xff;
	g1 = ((*src)>>8) & 0xff;
	r1 = ((*src)>>16) & 0xff;
	(*dest) = /* Y0 */ (((0x00100000 + r * IM11 + g * IM12 + b * IM13)>>16)&0x000000ff)
	        | /* U0 */ (((0x00800000 + r * IM31 + g * IM32 + b * IM33)>> 8)&0x0000ff00)
		| /* V0 */ (((0x00800000 + r * IM21 + g * IM22 + b * IM23)<< 8)&0xff000000)
		| /* Y1 */ (((0x00100000 + r1* IM11 + g1* IM12 + b1* IM13)    )&0x00ff0000)
		;
	dest++;
	src++;
 	}
}

// include assembly rgb2yuv routines:
#define rgb2yuv_asm
#include "x86_asm.h"
#undef rgb2yuv_asm


// ConvertRGB2YUV_MEM - convers [count] pixels from [*src] to [*dest] using bigarray
// transformation is FROM RGB colorspace TO YUY2 packed format
// This should be portable. Consumes hell a lot of memory (64MB). Using it on a unitialized array could be catastrophic
// Really NO precision loss - always correct results.
// count must be even
void ConvertRGB2YUV_MEM(Uint32 *dest, Uint32 *src, size_t count)
{
 count /= 2;
 while (count--) {
 	(*dest++) = bigarray[(*src)&0xffffff] + ((bigarray[(*(src+1))&0xffffff]<<16)&0xff0000);
	src+=2;
 	}
}


// Master function, switches to the fastest available conversion routine
/***
 *** you may wonder why I've included the X86 version at all (it'll be always slower than the X86_ASM version).
 *** well, someday this thing would be really portable ... hopefully...
 *************************************************************************************************************/

void ConvertRGB2YUV(Uint32 *dest, Uint32 *src, size_t count)
{
 switch (bestmethod) {
 	case USE_X86: ConvertRGB2YUV_X86(dest, src, count); break;
	case USE_MMX: ConvertRGB2YUV_MMX(dest, src, count); break;
	case USE_X86_ASM: ConvertRGB2YUV_X86_ASM(dest, src, count); break;
	case USE_X86_FPU: ConvertRGB2YUV_X86_FPU(dest, src, count); break;
	case USE_MMX2: ConvertRGB2YUV_MMX2(dest, src, count); break;
	case USE_SSE: ConvertRGB2YUV_SSE(dest, src, count); break;
	case USE_BIGMEM: ConvertRGB2YUV_MEM(dest, src, count); break;
 	}
}


// full YUV conversion ... when we precalculate we could be as slow (but precise) as hell :-)
static inline Uint32 fRGB2YUV(int r, int g, int b)
{int Y, U, V;
 Y = /*clamp(*/(int) (16+ (M11 * r + M12 * g + M13 * b))/*)*/;
 V = /*clamp(*/(int) (128+(M21 * r + M22 * g + M23 * b))/*)*/;
 U = /*clamp(*/(int) (128+(M31 * r + M32 * g + M33 * b))/*)*/;
 return (Y | (U<<8) | (V<<24));
}

// some string funcions:

// returns the position of the first occurence of PATTERN in S. If PATTERN cannot be found within S, pos returns -1
int pos(const char *pattern, const char *s)
{unsigned int j, i, ll, f;

 ll = strlen(pattern);
 for (j=0;j<strlen(s);j++) {
	f = 1;
	for (i=0;i<ll;i++)
		if (pattern[i]!=s[j+i]) {
			f = 0; break;
			}
	if (f) return j;
 	}
 return -1;
}

/***************************************************************
 * void GetCPUCaps - returns some flags about the processor's  *
 *                   capabilities                              *
 *-------------------------------------------------------------*
 *                                                             *
 * The Linux version relies heavily on the /proc filesystem :) *
 ***************************************************************/
#ifdef linux
void GetCPUCaps(CPUCaps *a)
{FILE *f;
 int i, k, memtotal;
 char buf[256];

// nullificate the CPU capabilities:
 a->hasmmx = 0;
 a->hasmmx2= 0;
 a->hassse = 0;
 a->largemem=0;
#ifndef USE_ASSEMBLY
 return;
#endif
 if (OptionExists("--has-mmx")) a->hasmmx=1;
 if (OptionExists("--has-sse")) a->hassse=1;
 if (OptionExists("--has-mmx2")) a->hasmmx2=1;
 if ((f=fopen("/proc/cpuinfo", "rt"))==NULL) {
 	printf("Can't open /proc/cpuinfo... falling back on defaut (hardcoded) config (MMX=0, BIGMEM=0)\n");
	return;
 	}
while (!feof(f)) {
	fgets(buf, 256, f);
	if (pos("flags", buf)!=-1) {
		// scan for cpu features
		if (pos("mmx", buf)!=-1) // do we have MMX
			a->hasmmx = 1;
		if (pos("sse", buf)!=-1) {// do we have SIMD instructions?
			a->hassse =  1;
			a->hasmmx2 = 1; //assume if we have SSE, we'll have MMX2...
		}
		if (pos("mmxext", buf)!=-1) // do we have the extended MMX instruction set?
			a->hasmmx2 =1;
		}
	}
 fclose(f);
 if (OptionExists("--no-mmx")) a->hasmmx=0;
 if (OptionExists("--no-sse")) a->hassse=0;
 if (OptionExists("--no-mmx2")) a->hasmmx2=0;
 if ((f=fopen("/proc/meminfo", "rt"))==NULL) {
 	printf("Can't get memory info - assuming no large memory allocation.");
	a->largemem = 0;
	return;
 	}
 while (!feof(f)) {
 	fgets(buf, 256, f);
	if (pos("MemTotal", buf)) {
		i = 0;
		while (buf[i]<'0'||buf[i]>'9') i++;
		k = i;
		while (buf[k]>='0'&&buf[k]<='9') k++;
		buf[k] = 0;
		sscanf(buf+i, "%d", &memtotal);
		// fixme : assuming memory is given in kilobytes:
		memtotal /= 1024;
		a->largemem = (memtotal >= MEMORY_REQUIREMENT);
		}
 	}
 fclose(f);
 if (OptionExists("--no-mem")) a->largemem = 0;
}
#else
#define rgbhacks1
#include "x86_asm.h"
#undef rgbhacks1
#endif
#define rgbhacks2
#include "x86_asm.h"
#undef rgbhacks2

// Benchmarks a RGB-to-YUY2 conversion function.
// Exact results may depend on the test data (fakeRGBBuffer) and the BENCHSIZE
// random data is a killer for ConvertRGB2YUV_MEM, as it performs about 6-10 times worse, but works fine on real data.
void BenchmarkFunction( __convert_fn_t fu, int code, char *method, int *maxfps, Uint32 timetorun)
{int cfps, frames=0;
 Uint32 ticks;
 printf("  ConvertRGB2YUV%s: ", method);
 fu(fakeYUVBuffer, fakeRGBBuffer, BENCHSIZE);
 ticks = GetTicks();
 while (ticks == GetTicks())
 	/*nothing*/;
 ticks = GetTicks();
 while (GetTicks()-ticks<timetorun) {
 	fu(fakeYUVBuffer, fakeRGBBuffer, BENCHSIZE);
 	frames++;
 	}
 ticks = GetTicks() - ticks;
 cfps = frames * 1000 / ticks;
 printf("%d fps\n", cfps);
 if (cfps > *maxfps) { *maxfps = cfps; bestmethod = code; }
 fflush(stdout);
}

/*
	This is called only when RGB2YUV conversion is needed.
	The benchmark flag is just passed to yuv_benchmark and defines the
	amount of time each function is benchmarked.
*/
void init_yuv_convert(int benchmark)
{int i;
 RawImg bm;
 bm.data = NULL;
 // get CPU capabilities
 GetCPUCaps(&CPU);
 if (OptionExists("--force-mem"))
 	CPU.largemem = 1;
 // disable all benchmarking and algorithm selection if we have some input on the command line
 if (	OptionExists("--use-x86")    ||OptionExists("--use-x86asm")||
	OptionExists("--use-mmx")    ||OptionExists("--use-mem")   ||
	OptionExists("--use-x86-fpu")||OptionExists("--use-sse")   ||
	OptionExists("--use-mmx2")) {
 	bestmethod = -1;
	if (OptionExists("--use-x86")) bestmethod = USE_X86;
#ifdef USE_ASSEMBLY
	if (OptionExists("--use-x86asm")) bestmethod = USE_X86_ASM;
	if (OptionExists("--use-x86-fpu")) bestmethod = USE_X86_FPU;
#endif
	if (OptionExists("--use-mmx") && CPU.hasmmx) bestmethod = USE_MMX;
	if (OptionExists("--use-sse") && CPU.hasmmx2 && CPU.hassse) bestmethod = USE_SSE;
	if (OptionExists("--use-mmx2") && CPU.hasmmx2) bestmethod = USE_MMX2;
	if (bestmethod!=-1) return;
 	}

 // put random data in the RGB buffer... this will produce more realistic results when benchmarking
 // all convesion functions that utilize lookup tables (will eliminate cache on large tables)
 for (i=0;i<BENCHSIZE;i++)
 	fakeRGBBuffer[i] = rand()%0x1000000;
 // it would be better if we can load a pre-rendered image to test the bigarray method on some 'real' data...
 if (LoadBmp("data/fract_benchmark.bmp", &bm)) {
	memcpy(fakeRGBBuffer, bm.data, bm.x*bm.y*4);
 	FreeRaw(&bm);
 	}
 // if we have enough memory, then create the big lookup array.
 if (CPU.largemem) {
 	if ((bigarray=(Uint32 *) malloc(0x4000000))==NULL) {
		printf("Cannot allocate memory for fast lookup table! Won't use this method!\n");
		CPU.largemem = 0;
		} else
		{
		printf("precalculating bigarray...");
		fflush(stdout);
		for (i=0;i<0x1000000;i++)
			bigarray[i] = fRGB2YUV((i & 0xff0000)>>16, (i & 0xff00) >> 8, (i & 0xff));
		printf("done\n");
		}

 	}
 yuv_benchmark(benchmark);
}

Uint32 get_correct_cksum(void)
{
	int os=1, iss;
	CPUCaps a;
	if (csc!=0xffffffff) return csc;
#ifdef linux
	os=0;
#endif
	GetCPUCaps(&a);
	iss = 0;
	if (a.hasmmx) iss++;
	if (a.hasmmx2) iss++;
	if (a.hassse) iss++;
	csc = checksums[is_intel()][os][iss];
	return csc;
}

void switch_rgbmethod(void)
{
	switch (bestmethod) {
		case USE_X86	: bestmethod = USE_X86_ASM;	break;
		case USE_X86_ASM: bestmethod = USE_X86_FPU;	break;
		case USE_X86_FPU: bestmethod = USE_MMX;		break;
		case USE_MMX	: bestmethod = USE_MMX2;	break;
		case USE_MMX2	: bestmethod = USE_SSE;		break;
		case USE_SSE	: bestmethod = USE_X86;		break;
	}
}

void set_correct_cksum(Uint32 c)
{
    csc = c;
}

void rgb2yuv_close(void)
{
 if (bigarray)
 	free(bigarray);
}

// the parameter is to specify --benchmark mode. If set, all conversion functions are run one minute unconditionally.
void yuv_benchmark(int benchmark)
{int maxfps = 0;
 int timetorun;
 timetorun = (benchmark?BENCH_LARGE_TICKS:BENCHTICKS);
 printf("Benchmarking RGB-to-YUV conversion functions:\n");
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_X86, USE_X86, "_X86", &maxfps, timetorun);
#ifdef USE_ASSEMBLY
	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_X86_ASM, USE_X86_ASM, "_X86_ASM", &maxfps, timetorun);
	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_X86_FPU, USE_X86_FPU, "_X86_FPU", &maxfps, timetorun);
#endif	
 if (CPU.hasmmx)
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_MMX, USE_MMX, "_MMX", &maxfps, timetorun);
 if (CPU.hasmmx2)
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_MMX2, USE_MMX2, "_MMX2", &maxfps, timetorun);
 if (CPU.hasmmx2 && CPU.hassse) {
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_SSE, USE_SSE, "_SSE", &maxfps, timetorun);
	}
 if (CPU.largemem)
 	BenchmarkFunction((__convert_fn_t) ConvertRGB2YUV_MEM, USE_BIGMEM, "_MEM", &maxfps, timetorun);
 if (OptionExists("--use-mem") && CPU.largemem) bestmethod = USE_BIGMEM;
}
