/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 * Handles save / load procedures. Fract engine can load its state (camera *
 * position, spheres, animation, etc.) in text files (usually .fsv files)  *
 * and may also save its state as well.                                    *
 *                                                                         *
***************************************************************************/
#include <limits.h>
#include <string.h>
#include "MyGlobal.h"
#include "MyTypes.h"
#include "bitmap.h"
#include "sphere.h"
#include "saveload.h"

/* ------------------------------------ export section ------------------------------------ */
extern int spherecount;
extern sphere sp[];
extern int lx;
extern int ly;
extern int lz;
extern double cur[];
extern int Physics;
extern int CollDetect;
extern char *floor_texture;
extern char *default_font;
extern AniSphere AS[];
extern int anicnt;
extern double alpha, beta;
extern double sv_gravity, sv_air;
extern double fov;

/* ----------------------------------- variables ------------------------------------------ */

double cd[MAXCD][6];
int cd_frames;

double play_time;

#ifdef RECORD
int rec_initialized = 0, rec_init_try = 0;
FILE *rec_file;
#endif

void write_comments(FILE *f)
{
	fprintf(f, "#\n# This is a fract-generated save context file.\n# Every line starting with a pound sign is ignored; ");
	fprintf(f, "Don't remove the first line, it is important.\n#\n# All data can be edited.\n#\n");
}

void write_triplet(FILE *f, const char *prefix, double x, double y, double z)
{
	fprintf(f, "%s.x=%.5lf\n", prefix, x);
	fprintf(f, "%s.y=%.5lf\n", prefix, y);
	fprintf(f, "%s.z=%.5lf\n", prefix, z);
}

void write_triplet1(FILE *f, const char *prefix, double *a)
{
	write_triplet(f, prefix, a[0], a[1], a[2]);
}

void write_sphere(FILE *f, int num, sphere a)
{int i;
	fprintf(f, "Sphere[%d].x=%.5lf\n", num, a.x);
	fprintf(f, "Sphere[%d].y=%.5lf\n", num, a.y);
	fprintf(f, "Sphere[%d].z=%.5lf\n", num, a.z);
	fprintf(f, "Sphere[%d].xi=%.5lf\n", num, a.xi);
	fprintf(f, "Sphere[%d].yi=%.5lf\n", num, a.yi);
	fprintf(f, "Sphere[%d].zi=%.5lf\n", num, a.zi);
	fprintf(f, "Sphere[%d].d=%.5lf\n", num, a.d);
	fprintf(f, "Sphere[%d].mass=%.5lf\n", num, a.mass);
	fprintf(f, "Sphere[%d].refl=%.5lf\n", num, a.refl);
	fprintf(f, "Sphere[%d].r=%d\n", num, a.r);
	fprintf(f, "Sphere[%d].g=%d\n", num, a.g);
	fprintf(f, "Sphere[%d].b=%d\n", num, a.b);
	fprintf(f, "Sphere[%d].AniIndex=%d\n", num, a.AniIndex);
	if (a.flags & MAPPED)
	fprintf(f, "Sphere[%d].texture=\"%s\"\n", num, getTexFn(a.tex));
	for (i=0;i<31;i++)
		if (flag2str[i][0])
			fprintf(f, "Sphere[%d].flags.%s=%d\n", num, flag2str[i], (((1<<i) & a.flags)!=0));
}

void write_ani(FILE *f, int num, AniSphere a)
{
	fprintf(f, "AS[%d].cx=%.5lf\n", num, a.cx);
	fprintf(f, "AS[%d].cy=%.5lf\n", num, a.cy);
	fprintf(f, "AS[%d].cz=%.5lf\n", num, a.cz);
	fprintf(f, "AS[%d].radius=%.5lf\n", num, a.radius);
	fprintf(f, "AS[%d].alpha=%.5lf\n", num, a.alpha);
	fprintf(f, "AS[%d].beta=%.5lf\n", num, a.beta);
	fprintf(f, "AS[%d].speed=%.5lf\n", num, a.speed);
}
// saves the current context to the given text file. If it exists, it is overwritten.
int SaveContext(const char *fn)
{FILE *f;
 int i;
 if ((f = fopen(fn, "wt"))==NULL)
 	return 0;
 fprintf(f, "%s\n", SAVEFILE_MAGIC);
 write_comments(f);
 fprintf(f, "\n##\n#Global Entries:\n#\n\n");
 fprintf(f, "\n# Light source coordinates:\n");
 write_triplet(f, "Light", lx, ly, lz);
 fprintf(f, "\n# User position information:\n");
 write_triplet1(f, "User", cur);
 fprintf(f, "alpha=%.6lf\n", alpha);
 fprintf(f, "beta=%.6lf\n", beta);
 fprintf(f, "\n# Physics (1=on)\n");
 fprintf(f, "Physics=%d\n", Physics);
 fprintf(f, "\n# Collision Detection (1=on)\n");
 fprintf(f, "CollDetect=%d\n", CollDetect);
 fprintf(f, "\n# Physics engine parameters\n");
 fprintf(f, "sv_gravity=%.6lf\n", sv_gravity);
 fprintf(f, "sv_air=%.6lf\n", sv_air);
 fprintf(f, "fov=%.6lf\n", fov);
 fprintf(f, "\n# Some vital file names\n");
 fprintf(f, "floor_texture=\"%s\"\n", floor_texture);
 fprintf(f, "default_font=\"%s\"\n", default_font);
 fprintf(f, "\n#\n# Sphere information section\n#\n\n");
 fprintf(f, "spherecount=%d\n", spherecount);
 for (i=0;i<spherecount;i++) write_sphere(f, i, sp[i]);
 fprintf(f, "\n#\n# Animation structure information section\n#\n\n");
 fprintf(f, "anicnt=%d\n", anicnt);
 for (i=0;i<anicnt;i++) write_ani(f, i, AS[i]);
 fclose(f);
 return 1;
}

#ifdef RECORD
// Does a record snapshot - writes down the current values of cur[], alpha and beta along with the current time
void record_do(double time)
{
	// check if the work file should be opened first
	if (rec_init_try) {
		if (!rec_initialized) return;
		fprintf(rec_file, "%.3lf,\t%.6lf,\t%.6lf,\t%.6lf,\t%.6lf,\t%.6lf,\n", time, cur[0], cur[1], cur[2], alpha, beta);
		} else
		{
		rec_init_try = 1;
		rec_file = fopen(RECORD_FILE, "wt");
		rec_initialized = 0;
		if (rec_file!=NULL) {
			rec_initialized = 1;
			record_do(time); // recurse once.
			}
		}
}
#endif

// hash function, which operates on a single line: hashes everything to the first occurence
// of a '=' character (if it does not exist, the whole line is hashed)
static Uint16 hashid(char *s)
{
	register Uint32 acc;
	int i, l;
	char sc;

	l = 0;
	while (s[l]!=0 && s[l]!='=' && s[l]!='[' && s[l]!='.') l++;
	sc=s[l];
	s[l]=0;
	for (acc=0, i=0;i<(l&~1);i++)
		acc += (Uint32) s[i] + (((Uint32) s[i+1])<<8);
	if (l%2)
		acc ^= s[l-1];
	s[l]=sc;
	acc = (acc&0xffff)^(acc>>16);
	return acc;
}

static inline int  line_empty(char *s)
{
	int i;

	i = 0;
	while (s[i]!=0) {
		if (s[i]!='\n' && s[i]!='#' && s[i]!=' ' && s[i]!='\t') return 0;
		i++;
		}
	return 1;

}

static inline int get_int(char *s)
{
	double f;
	int i;

	for (i=0;s[i]!='=';i++);
	i++;
	if (sscanf(s+i, "%lf", &f)) {
		if (f < -INT_MAX || f > INT_MAX) {
			printf("Value at line `%s' won't fit in the integer type!\n", s);
			return 0;
		} else	return ((int) f);
		}
		else {	printf("Value not read well at line `%s'!\n", s);
			return 0;
			}
}

static inline double get_real(char *s)
{
	double f;
	int i;

	for (i=0;s[i]!='=';i++);
	i++;
	if (sscanf(s+i, "%lf", &f)) return f;
		else {	printf("Value not read well at line `%s'!\n", s);
			return 0.0;
			}
}

static void get_int_triple(char *s, int *x, int *y, int *z)
{
	int i=0;

	while (s[i]!='.') i++;
	i++;
	s += i;
	if (s[0]=='x') *x = get_int(s);
	if (s[0]=='y') *y = get_int(s);
	if (s[0]=='z') *z = get_int(s);
}

static void get_real_triple(char *s, double *x, double *y, double *z)
{
	int i=0;

	while (s[i]!='.') i++;
	i++;
	s += i;
	if (s[0]=='x') *x = get_real(s);
	if (s[0]=='y') *y = get_real(s);
	if (s[0]=='z') *z = get_real(s);
}

static inline void get_str(char *s, char *so)
{
	int i=0,j;
	while (s[i]!=0 && s[i]!='=') i++;
	if (s[i]==0) { so[0]=0; return; }
	j = ++i;
	while (s[j]!=0) j++;
	j--;
	if (s[j]=='\n') s[j]=0;
	strcpy(so, s+i);
}

static void get_sphere_flags(char *s, int *f)
{
	int bittst;
	int i, j;

	j = 0;
	while (s[j]!='=') j++;
	s[j]=0;
	for (i=0,bittst=1ul;i<32;i++,bittst*=2)
		if (!strcmp(s, flag2str[i]))
			if (s[j+1]=='1')
				*f |= bittst;
			else
				*f &= ~bittst;
}

static void get_sphere_info(char *si)
{
	int i, j, index;
	char *s;

	for (i=0;si[i]!='[';i++);
	j = ++i;
	while (si[j]!=']') j++;
	si[j]=0;
	sscanf(si+i, "%d", &index);
	i = j+2; // after the dot
	s = si+i;
	switch (hashid(s)) {
		case 0x0078: sp[index].x	=	get_real(s); break;
		case 0x0079: sp[index].y	=	get_real(s); break;
		case 0x007a: sp[index].z	=	get_real(s); break;
		case 0x69e1: sp[index].xi	=	get_real(s); break;
		case 0x69e2: sp[index].yi	=	get_real(s); break;
		case 0x69e3: sp[index].zi	=	get_real(s); break;
		case 0x0064: sp[index].d	=	get_real(s); break;
		case 0x48b5: sp[index].mass	=	get_real(s); break;
		case 0x38a8: sp[index].refl	=	get_real(s); break;
		case 0x8cfb: sp[index].opacity	=	get_real(s); break;
		case 0x0072: sp[index].r	=	get_int (s); break;
		case 0x0067: sp[index].g	=	get_int (s); break;
		case 0x0062: sp[index].b	=	get_int (s); break;
		case 0xd212: sp[index].AniIndex =	get_int (s); break;
		case 0xa8e8: get_sphere_flags(s+6, &sp[index].flags); break;
		default:
			printf("LoadContext: The following line has no meaning to me:\n[%s]\n", s); fflush(stdout);
			break;
		}
}

static void get_ani_info(char *si)
{
	int i, j, index;
	char *s;

	s = si;
	for (i=0;s[i]!='[';i++);
	j = ++i;
	while (s[j]!=']') j++;
	s[j]= 0;
	sscanf(s+i, "%d", &index);
	j+=2;
	s = si + j;

	switch (hashid(s)) {
		case 0x78db: AS[index].cx     = get_real(s); break;
		case 0x79dc: AS[index].cy     = get_real(s); break;
		case 0x7add: AS[index].cz     = get_real(s); break;
		case 0x188a: AS[index].radius = get_real(s); break;
		case 0xa6c5: AS[index].alpha  = get_real(s); break;
		case 0x3b9d: AS[index].beta   = get_real(s); break;
		case 0x9fc8: AS[index].speed  = get_real(s); break;
		default:
			printf("LoadContext: The following line has no meaning to me:\n[%s]\n", s); fflush(stdout);
			break;
		}
}

// gets a context from a fract-written text file. On failure, returns 0.
// Loads everything (including textures)
int LoadContext(const char *fn)
{FILE *f;
 char line[256];
 int shb = 0;

 int input_type = 0;

 cd_frames = 0;
 if ((f = fopen(fn, "rt")) == NULL) {
 	printf("LoadContext: Can't open context file `%s'!\n", fn);
	return 0;
 	}
 fgets(line, 256, f);
 if (strcmp(line, SAVEFILE_MAGICn)) {
 	printf("LoadContext: Can't open context file [%s]: bad magic!\n", fn);
 	return 0;
	}
 while (!feof(f)) {
 	fgets(line, 256, f);
	if (!line_empty(line) && line[0]!='#') {
		/*printf("%s\n", line);
		fflush(stdout);*/
		switch (hashid(line)) {
		/* LIGHT */
			case 0xadf1: get_int_triple (line, &lx, &ly, &lz); break;
		/* CAMERA */
			case 0x4b9e: get_real_triple(line, cur+0, cur+1, cur+2); break;
			case 0xa6c5: alpha = get_real(line); break;
			case 0x3b9d: beta = get_real(line); break;
		/* Physics: */
			case 0x9501: Physics = get_int (line); break;
			case 0xa3e0: CollDetect = get_int (line); break;
			case 0xdf4d: sv_gravity = get_real(line); break;
			case 0x1386: sv_air = get_real(line); break;
			case 0xe5a3: fov = get_real(line); break;
		/* Textures: */
			case 0x314d: /*get_str (line, floor_texture); */break;
			case 0x9bff: /*get_str (line, default_font);*/ break;
		/* Spheres: */
			case 0x414c: spherecount = get_int (line); break;
			case 0x1665: get_sphere_info(line); break;
		/* Ani structures: */
			case 0x1e7f: anicnt = get_int (line); break;
			case 0x5394: get_ani_info(line); break;
		/* MISC: */
			case 0x2988: input_type = SAVLOAD_COORDS; shb = 1; break;
			default:
				printf("LoadContext: The following line has no meaning to me:\n[%s]\n", line); fflush(stdout);
				break;
			}
		}

	if (shb) break;
 	}
 if (input_type == SAVLOAD_COORDS) {
 	cd_frames = 0;
 	while (fscanf(f, "%lf%lf%lf%lf%lf%lf", 	cd[cd_frames]  , cd[cd_frames]+1, cd[cd_frames]+2,
						cd[cd_frames]+3, cd[cd_frames]+4, cd[cd_frames]+5)==6) cd_frames++;
 	}
 fclose(f);
 if (input_type)
 	return (input_type);
	else
	return (SAVLOAD_OK);
}

void load_frame(int frame_no)
{
	int i;
	if (frame_no>=cd_frames) return;
	play_time = cd[frame_no][0];
	for (i=0;i<3;i++) cur[i] = cd[frame_no][1+i];
	alpha = cd[frame_no][4];
	beta  = cd[frame_no][5];
}

void saveload_close(void)
{
#ifdef RECORD
	if (rec_init_done)
		fclose(rec_file);
#endif
}
