/***************************************************************************
 *   Copyright (C) 2004 by Veselin Georgiev                                *
 *   vesko@ViruS                                                           *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *-------------------------------------------------------------------------*
 *                                                                         *
 *      VECTORMATH.CPP - Does anything related to vector calculation:      *
 *        - sphere pre-projection                                          *
 *        - phong shading                                                  *
 *        - physics engine                                                 *
 *                                                                         *
 ***************************************************************************/

#include "MyGlobal.h"
#include "MyTypes.h"
#include "sphere.h"
#include "profile.h"
#include "gfx.h"
#include "vectormath.h"
#include "render.h"
#include <math.h>

#ifdef _MSC_VER
// 1/Pi
#define M_1_PI 0.31830988618379067153776752674503
#endif
// this should be 45 degrees, or M_PI/4
#define A_ALPHA_OFFSET 0.78539816339745
// this should be 3/4 of A_ALPHA_OFFSET
#define A_BETA_OFFSET  0.58904862254809

// this is obtained as the refraction factor Ni/Nr, where Ni is the refraction index of the air (1.0002926), and Nr is
// currently 1.51714, which is the refraction index of the glass
#define REFRACT_FACTOR 0.65932781417667

// this is the square of the refract factor
#define REFRACT_FACTOR_SQR 0.43471316654699
// 1 - REFRACT_FACTOR_SQR
#define REFRACT_FACTOR_SQR_M1 0.56528683345301

//#define REFRACT_FACTOR 0.9
//#define REFRACT_FACTOR_SQR 0.81
//#define REFRACT_FACTOR_SQR_M1 0.19

#define AIRTHICKNESS_NORMAL 0.5
#define AIRTHICKNESS_HIGH   2.5
#define YI_THRESH 1

#define TIME_EPS 0.00025
// this should be the reciprocal of TIME_EPS
#define T_1_TIME_EPS 4000.0


// how much of the energy is preserved in a bounce (1 means that bounce takes no energy)
#define BOUNCE_COEFF 0.666

/*double ambient = 0.2;
double diffuse = 0.55;
double specular = 0.15;*/
extern sphere sp[];
extern int spherecount;
extern double fov;
extern double daFloor, daCeiling;
extern int vframe;
extern int ysqrd_floor, ysqrd_ceil, ysqrd_raytrace;

int CollDetect = 1;
int Physics = 1;
//double sv_gravity = 166.66;
double sv_gravity = 0.0;
double sv_air     = 3.99;

double lix[3];

unsigned char m0ving[MAX_SPHERES];



#define InitMatrix(m) m[0][0]=a[0], m[1][0]=b[0], m[2][0]=-c[0], m[0][1]=0, m[1][1]=b[1], m[2][1]=-c[1], m[0][2]=a[2], m[1][2]=b[2], m[2][2]=-c[2]
#define InitMatrix2(m) m[0][0]=xv[0], m[0][1]=xv[1], m[0][2]=xv[2], m[1][0]=yv[0], m[1][1]=yv[1], m[1][2]=yv[2], m[2][0]=zv[0], m[2][1]=zv[1], m[2][2]=zv[2]
#define InitMatrix3(m) m[0][0]=a[0], m[0][1]=b[0], m[0][2]=-c[0], m[1][0]=a[1], m[1][1]=b[1], m[1][2]=-c[1], m[2][0]=a[2], m[2][1]=b[2], m[2][2]=-c[2]

#define ReplaceCol(m, col, a) m[col][0]=a[0], m[col][1]=a[1], m[col][2]=a[2]
#define ReplaceCol1(m,col, a) m[0][col]=a[0], m[1][col]=a[1], m[2][col]=a[2]
#define v1(a, b, c) c[0]=a[0]-b[0], c[1]=a[1]-b[1], c[2]=a[2]-b[2]
#define v2(a, b, c) c[0]=a[0]+b[0], c[1]=a[1]+b[1], c[2]=a[2]+b[2]
#define limit1(a) (a>1?1:(a))
#define arccos(x) (atan2(sqrt(1-(x)*(x)), x))
#define lift(x) ((x)*(x))
#define eq(a,b) (iabs((a)-(b))<EPS)


// calculates some grid point :
// INPUT:
// c -> camera position
// a - alpha of the point
// b - beta of the point
// alphaorig - alpha of the direction of the camera
// w - destination 3dpoint
void calc_grid_point(double *c, double a, double b, double alphaorig, double *w)
{
 w[0] = c[0] + sin(a) + sin(alphaorig)*(cos(b)-1);
 w[1] = c[1] + sin(b);
 w[2] = c[2] + cos(a) + cos(alphaorig)*(cos(b)-1);
}

// calculates the three ending points of the vector grid given a user viewpoint and a direction
void calc_grid_basics(double *c, double alpha, double beta, double w[3][3])
{
 calc_grid_point(c, alpha-A_ALPHA_OFFSET*fov*FOV_CORRECTION, beta+A_BETA_OFFSET*fov*FOV_CORRECTION, alpha, w[0]);
 calc_grid_point(c, alpha+A_ALPHA_OFFSET*fov*FOV_CORRECTION, beta+A_BETA_OFFSET*fov*FOV_CORRECTION, alpha, w[1]);
 calc_grid_point(c, alpha-A_ALPHA_OFFSET*fov*FOV_CORRECTION, beta-A_BETA_OFFSET*fov*FOV_CORRECTION, alpha, w[2]);
}

// calculates the determinant of the 3*3 matrix M.
double MatrixRoot(double M[3][3])
{
 return (M[0][0]*M[1][1]*M[2][2] + M[1][0]*M[2][1]*M[0][2] + M[0][1]*M[1][2]*M[2][0]
        -M[0][2]*M[1][1]*M[2][0] - M[2][2]*M[0][1]*M[1][0] - M[1][2]*M[2][1]*M[0][0]);
}

// determines where the center of the given sphere will map on the 2d screen
// works correctly with any FOV, X/Y ratio, etc.
// returns 0 if the center will be wildly of the screen (most probably when the sphere will not render at all)
int  ProjectSphere(sphere *d, double *cur, double w[3][3], int *dx, int *dy, int xres, int yres)
{
 double a[3], b[3], c[3], h[3];
 double m[3][3];
 double Dcr;
 double beta, gamma, delta;

 // init the a, b, c & h vectors
 v1(cur, w[0], h);
 v1(w[1], w[0], a);
 v1(w[2], w[0], b);
 c[0] = d->x - cur[0];
 c[1] = d->y - cur[1];
 c[2] = d->z - cur[2];

 InitMatrix(m);
 Dcr = MatrixRoot(m);
 if (iabs(Dcr)<0.01) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - zero determinant
	}

 // solve the ugly equation
 //InitMatrix(m);
 ReplaceCol(m, 0, h);
 beta = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 1, h);
 gamma = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 2, h);
 delta = MatrixRoot(m) / Dcr;

 if (delta < 0) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - behind user
	}
 if (beta < -0.2 || beta > 1.2) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - out of screen
	}
 if (gamma <-0.2 || gamma> 1.2) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - out of screen
	}
 *dx = (int) round(beta * xres);
 *dy = (int) round(gamma* yres);
 return 1;
 }

// determines where the given point will map on the 2d screen
// works correctly with any FOV, X/Y ratio, etc.
// returns 0 if the point is behind the user and won't render at all (which holds even for fov = 180).
int  ProjectPoint(int *dx, int *dy, double *d, double *cur, double w[3][3], int xres, int yres)
{
 double a[3], b[3], c[3], h[3];
 double m[3][3];
 double Dcr;
 double beta, gamma, delta;

 *dx=0; *dy=0;
 // init the a, b, c & h vectors
 v1(cur, w[0], h);
 v1(w[1], w[0], a);
 v1(w[2], w[0], b);
 v1(d, cur, c);

 InitMatrix(m);
 Dcr = MatrixRoot(m);
 if (iabs(Dcr)<0.01) {
 	//printf("Negative Determinant - ");
 	return 0; // sphere is inprojectable - zero determinant
	}

 // solve the ugly equation
 //InitMatrix(m);
 ReplaceCol(m, 0, h);
 beta = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 1, h);
 gamma = MatrixRoot(m) / Dcr;
 InitMatrix(m);
 ReplaceCol(m, 2, h);
 delta = MatrixRoot(m) / Dcr;

 if (delta < 0) {
 	//printf("bad delta");
 	return 0; // sphere is inprojectable - behind user
	}
 *dx = (int) round(beta * xres);
 *dy = (int) round(gamma* yres);
 return 1;
 }

// Projects where the shadow of the center of the sphere will map on the screen.
// Returns 0 if that point is out of screen or the light projection goes to infinity (with an
// epsilon DST_THRESHOLD, used for floor rendering)
// returns 1 if the point is mapped correctly
int  ProjectSphereShadow(sphere *s, double *l, int *x_2d, int *y_2d, int xres, int yres, double *cur, double *tt, double *a, double *b)
{
	double x, y, y_to_be, z, planeDist, m;
	double M[3][3], c[3], h[3], lambda, theta, zurla, Dcr;
	y = s->y - l[1];
	if (fabs(y) < DST_THRESHOLD) return 0; // sphere shadow goes to infinity
	if (y < 0.0) {
		planeDist = l[1] - daFloor;
		y_to_be = daFloor;
	} else {
		planeDist = daCeiling - l[1];
		y_to_be = daCeiling;
	}

	m = planeDist / fabs(y);

	x = l[0] + (s->x - l[0])*m;
	z = l[2] + (s->z - l[2])*m;
	y = y_to_be;

	// we're done with 3D projection. The sphere hits floor/ceiling in (x, y, z);

	// the following is based on the equation:
	/*
		  P = tt + lamda * ti + theta * tti = c + zurla * ((x,y,z)-c)

		the equaton is only one, but since we solve it for x, y and z, we get the three
		unknowns (lambda, theta & zurla)
	*/
	c[0] = x - cur[0]; c[1] = y - cur[1]; c[2] = z - cur[2];
	v1(cur, tt, h); // h = cur - tt -> vector from tt to cur
	InitMatrix3(M);

	Dcr = MatrixRoot(M);
	if (fabs(Dcr)<1E-8) {
		printf("Cannot solve shadow equation, zero determinant!\n");
		return 0;
	}

	ReplaceCol1(M, 0, h);
	lambda = MatrixRoot(M) / Dcr;
	if (lambda < 0.0 || lambda >= xres) return 0;
	*x_2d = (int) lambda;
	InitMatrix3(M);
	ReplaceCol1(M, 1, h);
	theta = MatrixRoot(M) / Dcr;
	if (theta < 0.0 || theta >= yres) return 0;
	*y_2d = (int) theta;
	InitMatrix3(M);
	ReplaceCol1(M, 2, h);
	zurla = MatrixRoot(M) / Dcr;
//	if (zurla < 0.0) return 0;
//	return 1;
	return (zurla>0.0);
}

double predlo[MAX_SPHERES], dlc;

void PassPre(double *p, double plc, double *LL)
{int i;
 dlc = plc;
 for (i=0;i<MAX_SPHERES;i++) predlo[i] = p[i];
 for (i=0;i<3;i++) lix[i]=LL[i];
}

// normalizes a vector at a, so that its length would become 1.
void norm(double *a)
{double t;
 t = 1 / sqrt( a[0]*a[0] + a[1]*a[1] + a[2]*a[2] );
 a[0]*=t; a[1]*=t; a[2]*=t;
}

/*static double ttz[MAX_CPU_COUNT];
static double determinant[MAX_CPU_COUNT];
static double gB[MAX_CPU_COUNT];*/

/*int testsphere(sphere *a, double *v, double *cur, double A)
{
 return ( (ttz=(sqr(a->x-(cur[0]+v[0]*a->dist)) + sqr(a->y-(cur[1]+v[1]*a->dist)) + sqr(a->z-(cur[2]+v[2]*a->dist))))<=sqr(a->d) );
}*/
// returns 1 if the vector v (originating from cur) hits the sphere a
// returns 0 otherwise
int testsphere1(sphere *a, double *v, double *cur, double A, double *determinant, double *gB)
{
 double C, x, y, z;
 x = cur[0] - a->x;
 y = cur[1] - a->y;
 z = cur[2] - a->z;
 *gB = x*v[0] + y*v[1] + z*v[2];
 C = x*x + y*y + z*z - a->d*a->d;
 *determinant = (*gB)*(*gB) - A*C;
 return (*determinant>=0.0f);
}

// normalizes an angle in the 0..2*pi range
static inline double normangle(double angle)
{
 angle *= M_1_PI;
 if (angle<0)
 	angle+=2;
 if (angle>2)
 	angle-=2;
 if (angle<0 || angle >2)
 	{printf("angle error!"); fflush(stdout);}
 return angle;
}


/*************************************************************************
 * search exact dist -- find the distance to the sphere's surface        *
 *                                                                       *
 * based on the fact, that there is always a factor Q, such that         *
 *                                                                       *
 * dist(sphere, cur + Q*v) = sphere.diameter                             *
 *                                                                       *
 * it yeilds to the following quadratic equation:                        *
 *                                                                       *
 * q^2*(vx^2+vy^2+vz^2)+2*q*(fx*vx+fy*vy+fz*vz)+fx^2+fy^2+fz^2-d.d^2 = 0 *
 * where f is a vector (cur - sphere.center), d.d is sphere radius       *
 *                                                                       *
 * There should be two solutions; we are interested in the smaller       *
 * of them only.                                                         *
 *                                                                       *
 *************************************************************************/
static inline double SearchExactDist(sphere *a, double *c, double *v)
{double f[3];
 double ua, ub, uc, Dt;
 f[0]= c[0]-a->x; f[1]=c[1]-a->y; f[2]=c[2]-a->z;
 ua = sqr(v[0]) + sqr(v[1]) + sqr(v[2]);
 ub = (f[0]*v[0] + f[1]*v[1] + f[2]*v[2])*2;
 uc = sqr(f[0]) + sqr(f[1]) + sqr(f[2]) - sqr(a->d);
 Dt = sqr(ub) - 4*ua*uc;
 //if (Dt<0) { printf("SearchExactDist:error: negative determinant! it should only be positive!!!\n"); fflush(stdout); Dt=0;}

 // return the smaller root of the equation
 return ((-ub-sqrt(Dt))/(2*ua));
}


/* FindNormalVector - finds the normal of the plane, defined by the points a, b and c.

   returns 0 if the math operation is unapplicable (that is, the points are colinear)
   returns 1 and a vector in v  otherwise
*/
static inline int FindNormalVector(double *a, double *b, double *c, double *v)
{double u[3], w[3];
 v1(b, a, u);
 v1(c, a, w);
 // are u and w colinear?
 if (eq(u[0]/w[0], u[1]/w[1]) && eq(u[1]/w[1], u[2]/w[2]))
 	return 0;
 // if they are not colinear, we can calculate the cross product of u and w, which will give us the normal of the plane
 v[0] = u[1]*w[2] - u[2]*w[1];
 v[1] = u[0]*w[2] - u[2]*w[0];
 v[2] = u[0]*w[1] - u[1]*w[0];
 return 1;
}



/**********************************************************************************
 *    SOLVE3D - calculates the phong shading for a point I, which is the          *
 *              crossing point of the sphere A and the vector V, originating      *
 *              from the camera (point C), with a light source of L.              *
 *              `iteration' is used for raytracing - pass 0 when calling Solve3D  *
 *                                                                                *
 *    Returns:                                                                    *
 *             - a color, depending on the calculated phong and the r, g and b    *
 *               values of the sphere                                             *
 *             - an opacity, double between 0.0 and 1.0. Opacity is 1 anywhere    *
 *               except for the edge of the sphere, where it gradually drops.     *
 *                                                                                *
 **********************************************************************************/
Uint32 Solve3D(sphere *a, double *v, double *c, double *l, double Arcp, float *opacity, double *determinant, double *gB, int iteration)
{double d[3], i[3], xv[3], f[3], e[3], li[3], rt[3], doi=1.0/a->d, dd, dci, lambda;
 double cosbeta, cosbetalim;
 double IE, CE2;
 int rezult;
 int i2x, i2y;
 Uint32 ttt;

 double IP, IOp, IOpv[3], Ip[3], Op[3], q[3], cosalpha, cosalphap, o[3], OOp;

#ifdef MAKE_PROFILING
 if (!iteration) prof_enter(PROF_SOLVE3D);
#endif

 *determinant = sqrt(*determinant); // <- this assumes we've called testsphere before Solve3D
 lambda = (-(*gB) - (*determinant)) * Arcp;
 if (!iteration) {	// check if we're close to the edge:
 	if (*determinant * doi > 0.25f) *opacity = 1.0f;
 			else    *opacity = *determinant*doi*4.0f;
 }

// d is the sphere's center
 d[0] = a->x; d[1] = a->y; d[2] = a->z;

 // calculate the point I where the ray from the user crosses the sphere's surface
 i[0] = c[0] + lambda * v[0];
 i[1] = c[1] + lambda * v[1];
 i[2] = c[2] + lambda * v[2];

 // find the distance from the viewer's point to the sphere
 dci = pith(i,c);

 // xv is a vector, originating from D, pointing towards I
 v1(i, d, xv);

 xv[0]*=doi; xv[1]*=doi; xv[2]*=doi; //normalise the vector xv

 // li is a vector, originating from I, pointing towards L
 v1(l, i, li);

 // find the dot product of li and xv
 dd = li[0]*xv[0] + li[1]*xv[1] + li[2]*xv[2];

 // point f is the projection of L onto the vector xv
 f[0] = i[0] + dd*xv[0];
 f[1] = i[1] + dd*xv[1];
 f[2] = i[2] + dd*xv[2];

 /**********************************
  a little bit of explanation here:

  we're working on a plane, defined by
  the points L and I and the vector xv.

  Point E is on the same plane and
  is a mirror image of L, with a mirroring
  line xv.

  **********************************/
 e[0] = l[0] + 2*(f[0] - l[0]);
 e[1] = l[1] + 2*(f[1] - l[1]);
 e[2] = l[2] + 2*(f[2] - l[2]);

 // here in E we have the 3dpoint of the reflection of L from the ball (the light ray goes L -> I -> E

 // to calculate phong shading we need the cosine of the angle CIE

 CE2 = sqr(c[0]-e[0]) + sqr(c[1] - e[1]) + sqr(c[2] - e[2]);
 IE = pith(i, e);
 cosbeta = limit1((sqr(dci) + sqr(IE) - CE2) / (2.0f*dci*IE));
 //if (cosbeta<-1) { printf("Bad beta - %.5lf\n"); fflush(stdout); }
 cosbetalim = sqrt( (1 + cosbeta) * 0.5f );
 // the original Phong Shading formulae:
 //cosbetalim += 0.005 / (cosbetalim+0.05);
 cosbetalim += 0.2f;
 rezult = (int) ((ambient + lift(cosbeta)*specular + cosbetalim * diffuse)*65536.0);
 rezult = (rezult>65536)?65536:rezult;
 if (a->flags & MAPPED) {
 	i2x = (int) (0.5*normangle(atan2(-i[0]+a->x, i[2]-a->z))*a->tex->x);
	i2y = (int) (0.5*((a->y+a->d-i[1])/a->d)*a->tex->y);
	rezult = multiplycolor(a->tex->data[(i2x + i2y * a->tex->x)%(a->tex->x*a->tex->y)], rezult);
 	}
 else rezult=(((a->b * rezult)>>16) + (((a->g * rezult)>>16)<<8) + ((a->r * rezult) & 0xff0000));
#ifdef MAKE_PROFILING
 if (!iteration) prof_leave(PROF_SOLVE3D);
#endif
 if (! (a->flags & RAYTRACED)) return rezult;
 	else {
#ifdef MAKE_PROFILING
		if (!iteration) prof_enter(PROF_RAYTRACE);
#endif
	 // li is a vector, originating at I, pointing towards C
 	v1(c, i, li);

 	// find the dot product of li and xv in dd
	li[0]*=xv[0]; li[1]*=xv[1]; dd = li[2] * xv[2];
	dd+= li[0]+li[1];

 	// point f is the projection of C onto the vector xv
 	f[0] = i[0] + dd*xv[0];
 	f[1] = i[1] + dd*xv[1];
 	f[2] = i[2] + dd*xv[2];

 	e[0] = c[0] + 2.0f*(f[0]-c[0]);
 	e[1] = c[1] + 2.0f*(f[1]-c[1]);
 	e[2] = c[2] + 2.0f*(f[2]-c[2]);

	v1(e, i, rt);

	ttt = blend(Raytrace(i,rt, a->flags & RECURSIVE, iteration+1, a), rezult, a->refl);
	if (a->flags & TRANSPARENT) { // is the sphere transparent?
		/*
			Explanation of the refraction algorithm:
			(see http://www.ps.missouri.edu/rickspage/refract/refraction.html)

				The user's ray comes from C and hits the sphere at I. The ray is then refracted.
			We have xv, the normalized vector from D to I (D is the sphere's centrum).
			The (lesser) angle between Xv and CI is called an incident angle. This is the angle Alpha.
			Let the vector CI continues (refracted) through the sphere until it hits the surface again in I'.
			According to the Snell's Law the incident angle Alpha' between II' and Xv can be calculated
			as follows:

			Sin(Alpha') = Ni/Nr*Sin(Alpha),
				where Ni and Nr are the Refraction Indices of the  materials, which the
			ray is leaving/entering correspondingly (that is, Ni is the RI(refractive Index) of the air, and
			Nr is the RI of the glass - see REFRACT_FACTOR define in the beginning of this source file)

			(Of course, we don't need the actual angle Alpha' (nor Alpha, btw), so we won't bother doing an arcsin
			to obtain the exact angle)

			Quick brief: We get Cos(Alpha) using dot prouct. Get Sin(Alpha) using Pithagor. Get Sin(Alpha') using
			Snell's Law. Get Cos(Alpha') using pithagor.

			Now we notice that, I, D and I' are on the same plane (let us denote that plane with Zeta). Zeta also
			contains the vector q, which exits the sphere after the second refraction. Let we denote the crossing point
			of the rays CI and q with O'. OO' cross II' = P. O'' belongs to OO' and O'P = PO''. Now, let the vector o
			denotes the vector O'O''. It is easy to see, that IOI'O' is a romboid in Zeta, and q = O'I' = IO' + o;
			moreover, II' = 2*IO' + o. As we have the exit point and the exit vector, we simply call raytrace to continue
			seeking what the ray "sees":)

			if you have the full source of the project, you may well consider looking in the refract_scheme.gif
		*/
		// cosine of alpha is the dot product of v and xv divided by the length of xv (1) * v(1)
		norm(v); //v is still not normalized... :)
		cosalpha = v[0]*xv[0] + v[1]*xv[1] + v[2]*xv[2];
		if (cosalpha<0) cosalpha = -cosalpha;
		cosalphap = sqrt(REFRACT_FACTOR_SQR_M1 + REFRACT_FACTOR_SQR*sqr(cosalpha));
		IP = a->d*cosalphap;

		IOp = IP / (cosalphap*cosalpha + REFRACT_FACTOR * (1 - sqr(cosalpha)));
		IOpv[0] = v[0] * IOp;
		IOpv[1] = v[1] * IOp;
		IOpv[2] = v[2] * IOp;
		Op[0] = i[0] + IOpv[0];
		Op[1] = i[1] + IOpv[1];
		Op[2] = i[2] + IOpv[2];
		// o is the vector, originating from Op, pointing towards d
		v1(d, Op, o);
		norm(o);
		OOp = sqr(IOp) - sqr(IP);
		if (OOp < 0) {//printf("OOp out of range!\n");
			OOp=0;
			}

		//OOp = 2.0 * sqrt(sqr(IOp) - sqr(IP));
		OOp = 2.0 * sqrt(OOp);
		o[0]*=OOp; o[1]*=OOp; o[2]*=OOp;
		Ip[0] = IOpv[0]*2.0 + o[0] + i[0];
		Ip[1] = IOpv[1]*2.0 + o[1] + i[1];
		Ip[2] = IOpv[2]*2.0 + o[2] + i[2];
		// we found I'...
		q[0] = IOpv[0] + o[0];
		q[1] = IOpv[1] + o[1];
		q[2] = IOpv[2] + o[2];

		// Ip is the exit point, q is the exit vector...
		ttt = blend(Raytrace(Ip,q , (a->flags & RECURSIVE) | RAYTRACE_BILINEAR_MASK, iteration+1, a), ttt, a->opacity);
	}
#ifdef MAKE_PROFILING
	if (!iteration) prof_leave(PROF_RAYTRACE);
#endif	
	return ttt;
	}
}

// returns the color of the ray v, originating from `cur'.
// `iteration' is a counter which prevents a ray from entering an endless cycle (very rare, but possible!)
// `recursive' is a boolean flag: if 0, the procedure won't check if the ray hits any sphere.
//                                if 1 and a sphere is being hit the same procedure is recursively called for the ray
//                                that goes out of the sphere then.
// `last_sphere' a pointer to the sphere the ray is traced from. This avoids a ray to hit again the sphere it just hit.
// `tex' is which mipmap level to use
Uint32 Raytrace(double *cur, double *v, int recursive, int iteration, sphere *last_sphere)
{double ydist,scalefactor;
 double cx, cy, dist, mind, gB, determinant;
 float dp;

 sphere *z;
 int i;
 if (iteration > MAX_RAYTRACE_ITERATIONS)
 	return RAYTRACE_BLEND_COLOR;
 if (recursive&RECURSIVE) {
 	norm(v);
	mind = 999999999.0;
	z = NULL;
	for (i=0;i<spherecount;i++) if (sp+i!=last_sphere) {
		if (testsphere1(sp+i, v, cur, 1.0,  &determinant, &gB)) {
			dist = (-gB - sqrt(determinant));
			if (dist >0.0f && dist < mind) {
				mind = dist;
				z = sp+i;
				}
			}
		}
	if (z!=NULL) {
		testsphere1(z, v, cur, 1.0, &determinant, &gB);
		return Solve3D(z, v, cur, lix, 1.0, &dp, &determinant, &gB, iteration+1);
		}
 	}
 ydist = fabs(v[1]);
 if (ydist<DST_THRESHOLD)
 	return RAYTRACE_BLEND_COLOR;
 if (v[1]<0.0) {
 	dp = cur[1]-daFloor;
	ysqrd_raytrace = ysqrd_floor;
 } else {
 	dp = daCeiling - cur[1];
	ysqrd_raytrace = ysqrd_ceil;
 }
 //dp = (v[1]<0?(cur[1]):(200-cur[1]));
 scalefactor = dp/ydist;
 cx = cur[0] + v[0]*scalefactor;
 cy = cur[2] + v[2]*scalefactor;
 return (recursive&RAYTRACE_BILINEAR_MASK)?texture_handle_bilinear(cx, cy):texture_handle_nearest((int) cx, (int) cy );
}


// performs time calculation.
//
//  time -> frame difference, in seconds
//  resistance -> air resistance, see below
//  value -> current speed

// resistance is:
// 0-1 for no resistance
// 2 - ball slows down to 50% of its speed for a second
// 9 - ball slows down to 1/9 of its speed for a second
// etc. etc.

// dir is direction. if 1 the direction is forward (air resisance is applied), otherwise is backward (resistance is UNapplied)
double time_calc(double time, double resistance, double value, int dir)
{
    double res;
    if (resistance <= 1.0) return value;
    if (dir == 1)
    	res = value / exp(log(resistance) * time);
	else
	res = value * exp(log(resistance) * time);
    return res;
}

// adds the gravity to the Y component of the vector
void apply_gravity(sphere *a, double t)
{
 if (a->flags & GRAVITY) {
 	a->yi -= sv_gravity*(t-a->time);
 }
}

// mask is a bitmask: which components will the air be applied to
// Use APPLY_ALL to mean all components
// dir is direction: if 1, the air resistance is applied, otherwise it is undone ("un-applied")
void apply_air(sphere *a, double t, int mask, int dir)
{
 double multiplyer;
 if (a->flags & AIR) {
 	if (a->flags & GRAVITY) multiplyer = sv_air * AIRTHICKNESS_NORMAL;
 			 else   multiplyer = sv_air * AIRTHICKNESS_HIGH;

 	if (mask & APPLY_X) a->xi = time_calc((t-a->time) , multiplyer, a->xi, dir);
 	if (mask & APPLY_Y) a->yi = time_calc((t-a->time) , multiplyer, a->yi, dir);
 	if (mask & APPLY_Z) a->zi = time_calc((t-a->time) , multiplyer, a->zi, dir);
 }
}


double get_hit_time(double a, double b, double c)
{
	double Dt;
	Dt = b*b - 4*a*c;
	if (Dt<0) {
		//printf("Bug: Determinant in get_hit_time < 0!\n");
		Dt = 0;
	}
	return (-b-sqrt(Dt)) / (2*a);
}

/********************************
 * void advance:                *
 *      input:                  *
 *      *a - sphere to advance; *
 *      t - sync time.          *
 *------------------------------*
 * moves a sphere depending on  *
 * its movement vector and the  *
 * time remaining to the sync   *
 * time. Checks if the sphere   *
 * hits the floor or the        *
 * ceiling - if so, a bounce is *
 * generated.                   *
 ********************************/

void advance(sphere *a, double t)
{double ytoadd, x;
 int bModif = 0;
 a->x += a->xi * (t-a->time);
 a->z += a->zi * (t-a->time);

 ytoadd = a->yi * (t-a->time);
 if (a->flags & GRAVITY)
 	if (a->y + ytoadd < daFloor + a->d) {// check for floor bounce
 		bModif = 1;
		/* so we found out that the bounce occurs. So we should make some correction to gravity, because
		 * `apply_gravity' assumes the acceleration towards the floor occurs all the way in the same direction.
		 * This ain't true. We should apply the gravity prior to the bounce, then change the sign of the Y component
		 * of the vector, and finally apply the needed amount of gravity after the bounce. (Thinking how gravity
		 * applies to the movement vector, it applies with different signs before and after the bounce)
		 */

		// first, we un-do the gravity, which has been previously applied by `apply_gravity'//
		// before that, undo air resistance:
		apply_air(a,t, APPLY_Y, -1);
		// undo gravity
		a->yi+= sv_gravity * (t - a->time);

		// calculate when it hits the floor.
		x = a->time + get_hit_time(-sv_gravity, a->yi, a->y - a->d);

		if (x>t) {
			//printf("Bug: bounce time was outside the sync time!\n");
			x = t;
			} else if (x<a->time) {
			//printf("Bug: bounce time was before start time!\n");
			x = a->time;
			}
		a->yi = -(a->yi - (x-a->time) * sv_gravity) * BOUNCE_COEFF;
		a->y = a->d + daFloor + (t-x)*a->yi;
		a->yi -= (t-x)*sv_gravity;

		if (a->yi<YI_THRESH) {
		// if the ball is slowed down enough, we may consider it stopped jumping and avoid
		// any further bounce checks
			a->yi=0;
			a->y = a->d+daFloor;
			a->flags &= ~GRAVITY;     // don't do gravity anymore.
			}
 		}
 if (a->flags & GRAVITY)
 	if (a->y + ytoadd > daCeiling - a->d) {// check for ceiling bounce
	// we use much simplier routine, since a ball cannot roll on the ceiling, right?
	// (except sv_gravity is below zero, but ... :)
 		bModif = 1;
		ytoadd = (a->y + ytoadd) - (daCeiling - a->d);
		a->y = daCeiling - a->d - ytoadd;
		a->yi = -a->yi * BOUNCE_COEFF;
 		}
 if (!bModif)
	a->y += ytoadd;
 a->time = t; // fix the ball's time to the sync
}

// inits the "m0ving" array: a sphere is considered moving iff its speed vector is more than EPS
void check_moving(void)
{
	int i;
	for (i=0; i < spherecount; i++)
		m0ving[i] = ( sqr(sp[i].xi) + sqr(sp[i].yi) + sqr(sp[i].zi) > MOVING_THRESH);
}

int can_collide(int i, int j)
{
	return (m0ving[i]||m0ving[j]);
}

// calculates the distance between a and b at time t
static inline double time_dist(sphere *a, sphere *b, double t)
{double t1, t2;
 t1 = t - a->time;
 t2 = t - b->time;
 return (sqrt(sqr(a->x+a->xi*t1-b->x-b->xi*t2)+sqr(a->y+a->yi*t1-b->y-b->yi*t2)+sqr(a->z+a->zi*t1-b->z-b->zi*t2)));
}

// checks if two spheres collide until time `t'.
// if collision occurs it returns the exact collision time. Else a value of -1 is returned
double collide (sphere *a, sphere *b, double t)
{double l, r, m, d0, d1;
 r = t;
 l = (a->time>b->time?a->time:b->time); // get the latter of the a's and b's times
 d0 = time_dist(a, b, l);
 while (r-l>TIME_EPS) {
 	m = (l+r)/2.0;
	d0 = time_dist(a, b, m);
	d1 = time_dist(a, b, m-TIME_EPS);
	if (d1 < d0)
		r = m;
	else
		l = m;
 	}
 if (d0 > a->d+b->d)
 	return -1; // the spheres do not collide
 //
#ifdef PRINT_COLLISIONS
 printf("Collision detected!\nA:\n", t);
 print_sphere_info(stdout, a);
 printf("B:\n");
 print_sphere_info(stdout, b);
 printf("--> lowest time : %.6lf\n", m);
 fflush(stdout);
#endif
 l = (a->time>b->time?a->time:b->time); // get the latter of the a's and b's times

 do {
 	m = (l+r) / 2.0;
	d0 = time_dist(a, b, m);
	if (d0 > a->d + b->d)
		l = m;
	else
		r = m;
 	} while (r-l > TIME_EPS);
#ifdef PRINT_COLLISIONS
 printf("--> collision time : %.6lf\n", l); fflush(stdout);
#endif
 return l;
}

// assumes that a and b collide at time t. Upon that assumption it calculates the new motion vectors for a and b
void ProcessIncident(sphere *a, sphere *b, double t)
{double strength, d1;
 a->x += a->xi * (t- a->time); a->y += a->yi * (t- a->time);  a->z += a->zi * (t- a->time);
 b->x += b->xi * (t- b->time); b->y += b->yi * (t- b->time);  b->z += b->zi * (t- b->time);
 a->time = t;
 b->time = t;
 d1 = time_dist(a, b, t);
 strength = d1 - time_dist(a, b, t+TIME_EPS);
 d1 *= 0.5;
 if (strength<0) return; // the hit is too weak
 if (!(a->flags & STATIC)) {
 	a->xi += (a->x - b->x)/d1 * b->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	a->yi += (a->y - b->y)/d1 * b->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	a->zi += (a->z - b->z)/d1 * b->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
	}
 if (!(b->flags & STATIC)) {
 	b->xi += (b->x - a->x)/d1 * a->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	b->yi += (b->y - a->y)/d1 * a->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
 	b->zi += (b->z - a->z)/d1 * a->mass / (a->mass+b->mass) * strength * T_1_TIME_EPS;
	}
 // custom onHit modification goes here:
 if (!(a->flags & STATIC)) {
 	a->flags &= ~ANIMATED;
	a->flags |= GRAVITY|AIR;
 	}
 if (!(b->flags & STATIC)) {
 	b->flags &= ~ANIMATED;
	b->flags |= GRAVITY|AIR;
 	}
}


void set_gravity(double new_gravity) {
	sv_gravity = new_gravity;
}

int ograve=0;

void switch_gravity(void)
{
	int i;
	double sign=1;
	ograve++;
	switch (ograve%4) {
		case 0:
		case 2: sv_gravity = 0.001; break;
		case 1: sign = -1;
		case 3: sv_gravity = -sign * 166.0;
			for (i=0;i<spherecount;i++)
				if (sp[i].flags & AIR)
					sp[i].flags |= GRAVITY;
			break;
	}

}

void vectormath_close(void)
{
 //printf("diffuse: %.2lf, specular: %.2lf\n", diffuse, specular);
}
